# -*- coding: utf-8 -*-


"""Molecular topological indices."""

from typing import Iterable, List

import numpy as np
from rdkit import Chem
from rdkit.Chem import GraphDescriptors as GD
from rdkit.Chem import rdchem

periodicTable = rdchem.GetPeriodicTable()


class Topology:
    """Topological indices."""

    @staticmethod
    def _get_principle_quantum_number(atNum: int) -> int:
        """Get principle quantum number from atomic number."""
        if atNum <= 2:
            return 1
        elif atNum <= 10:
            return 2
        elif atNum <= 18:
            return 3
        elif atNum <= 36:
            return 4
        elif atNum <= 54:
            return 5
        elif atNum <= 86:
            return 6
        else:
            return 7

    @staticmethod
    def calculate_weiner(mol: Chem.Mol) -> float:
        """Get Weiner values of a molecule.
    
        Or W.
        """
        return 1.0 / 2 * sum(sum(Chem.GetDistanceMatrix(mol)))

    @staticmethod
    def calculate_mean_weiner(mol: Chem.Mol) -> float:
        """Get Mean Weiner index of a molecule.
    
        Or AW.
        """
        N = mol.GetNumAtoms()
        WeinerNumber = Topology.calculate_weiner(mol)
        return 2.0 * WeinerNumber / (N * (N - 1))

    @staticmethod
    def calculate_balaban(mol: Chem.Mol) -> float:
        """Get Balaban index of a molecule.
    
        Or J.
        """
        adjMat = Chem.GetAdjacencyMatrix(mol)
        Distance = Chem.GetDistanceMatrix(mol)
        Nbond = mol.GetNumBonds()
        Natom = mol.GetNumAtoms()
        S = np.sum(Distance, axis=1)
        mu = Nbond - Natom + 1
        sumk = 0.
        for i in range(len(Distance)):
            si = S[i]
            for j in range(i, len(Distance)):
                if adjMat[i, j] == 1:
                    sumk += 1. / np.sqrt(si * S[j])
        if mu + 1 != 0:
            J = float(Nbond) / float(mu + 1) * sumk
        else:
            J = 0
        return J

    @staticmethod
    def calculate_graph_distance(mol: Chem.Mol) -> float:
        """Get graph distance index.
    
        Or Tigdi.
        """
        Distance = Chem.GetDistanceMatrix(mol)
        n = int(Distance.max())
        res = 0.0
        for i in range(n):
            temp = 1. / 2 * sum(sum(Distance == i + 1))
            res = res + temp ** 2
        return np.log10(res)

    @staticmethod
    def calculate_diameter(mol: Chem.Mol) -> float:
        """Get largest value of the distance matrix.
    
        Or diametert.
        From Petitjean, M. J. Chem. Inf. Comput. Sci. 1992, 32, 4, 331-337.
        """
        Distance = Chem.GetDistanceMatrix(mol)
        return Distance.max()

    @staticmethod
    def calculate_radius(mol: Chem.Mol) -> float:
        """Get radius based on topology.
    
        Or radiust.
        From Petitjean, M. J. Chem. Inf. Comput. Sci. 1992, 32, 4, 331-337.
        """
        Distance = Chem.GetDistanceMatrix(mol)
        temp = []
        for i in Distance:
            temp.append(max(i))
        return min(temp)

    @staticmethod
    def calculate_petitjean(mol: Chem.Mol) -> float:
        """Get Petitjean based on topology.
    
        Or petitjeant.
        """
        diameter = Topology.calculate_diameter(mol)
        radius = Topology.calculate_radius(mol)
        return (float(diameter) - radius) / radius

    @staticmethod
    def calculate_xu_index(mol: Chem.Mol) -> float:
        """Get Xu index.
    
        Or Xu.
        """
        nAT = mol.GetNumAtoms()
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        Distance = Chem.GetDistanceMatrix(mol)
        sigma = np.sum(Distance, axis=1)
        temp1 = 0.0
        temp2 = 0.0
        for i in range(nAT):
            temp1 = temp1 + deltas[i] * ((sigma[i]) ** 2)
            temp2 = temp2 + deltas[i] * (sigma[i])
        Xu = np.sqrt(nAT) * np.log(temp1 / temp2)
        return Xu

    @staticmethod
    def calculate_gutman_topo(mol: Chem.Mol) -> float:
        """Get Gutman molecular topological simple vertex index.
    
        Or GMTI.
        """
        nAT = mol.GetNumAtoms()
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        Distance = Chem.GetDistanceMatrix(mol)
        res = 0.0
        for i in range(nAT):
            for j in range(i + 1, nAT):
                res = res + deltas[i] * deltas[j] * Distance[i, j]
        return np.log10(res)

    @staticmethod
    def calculate_polarity_number(mol: Chem.Mol) -> float:
        """Get Polarity number.
    
        Or Pol.
        """
        Distance = Chem.GetDistanceMatrix(mol)
        res = 1. / 2 * sum(sum(Distance == 3))
        return res

    @staticmethod
    def calculate_pogliani_index(mol: Chem.Mol) -> float:
        """Get Poglicani index.
    
        Or DZ.
        From Pogliani L. J.Phys.Chem. (1996), 100,18065-18077.
        """
        res = 0.0
        for atom in mol.GetAtoms():
            n = atom.GetAtomicNum()
            nV = periodicTable.GetNOuterElecs(n)
            mP = Topology._get_principle_quantum_number(n)
            res = res + (nV + 0.0) / mP
        return res

    @staticmethod
    def calculate_ipc(mol: Chem.Mol) -> float:
        """Get Bonchev-Trinajstic complexity index.
    
        Or Ipc.
        From Bonchev D. & Trinajstic N., J. Chem. Phys. (1977) 67,4517-4533.
        """
        return np.log10(GD.Ipc(mol))

    @staticmethod
    def calculate_bertzct(mol: Chem.Mol) -> float:
        """Get Bertz complexity index.
    
        Or BertzCT.
        From Bertz S. H., J. Am. Chem. Soc. (1981) 103,3599-3601.
        """
        return np.log10(GD.BertzCT(mol))

    @staticmethod
    def calculate_harary(mol: Chem.Mol) -> float:
        """Get Harary number.
    
        Or Thara.
        """
        Distance = np.array(Chem.GetDistanceMatrix(mol), 'd')
        return 1.0 / 2 * (sum(1.0 / Distance[Distance != 0]))

    @staticmethod
    def calculate_schiultz(mol: Chem.Mol) -> float:
        """Get Schiultz number.
    
        Or Tsch.
        """
        Distance = np.array(Chem.GetDistanceMatrix(mol), 'd')
        Adjacent = np.array(Chem.GetAdjacencyMatrix(mol), 'd')
        VertexDegree = sum(Adjacent)
        return sum(np.dot((Distance + Adjacent), VertexDegree))

    @staticmethod
    def calculate_zagreb1(mol: Chem.Mol) -> float:
        """Get Zagreb index with order 1.
    
        Or ZM1.
        """
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        return sum(np.array(deltas) ** 2)

    @staticmethod
    def calculate_zagreb2(mol: Chem.Mol) -> float:
        """Get Zagreb index with order 2.
    
        Or ZM2.
        """
        ke = [x.GetBeginAtom().GetDegree() * x.GetEndAtom().GetDegree() for x in mol.GetBonds()]
        return sum(ke)

    @staticmethod
    def calculate_mzagreb1(mol: Chem.Mol) -> float:
        """Get Modified Zagreb index with order 1.
    
        Or MZM1.
        """
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        while 0 in deltas:
            deltas.remove(0)
        deltas = np.array(deltas, 'd')
        res = sum((1. / deltas) ** 2)
        return res

    @staticmethod
    def calculate_mzagreb2(mol: Chem.Mol) -> float:
        """Get Modified Zagreb index with order 2.
    
        Or MZM2.
        """
        cc = [x.GetBeginAtom().GetDegree() * x.GetEndAtom().GetDegree() for x in mol.GetBonds()]
        if len(cc) == 0:
            return 0.0
        while 0 in cc:
            cc.remove(0)
        cc = np.array(cc, 'd')
        res = sum((1. / cc) ** 2)
        return res

    @staticmethod
    def calculate_quadratic(mol: Chem.Mol) -> float:
        """Get Quadratic index.
    
        Or Qindex.
        """
        M = Topology.calculate_zagreb1(mol)
        N = mol.GetNumAtoms()
        return 3 - 2 * N + M / 2.0

    @staticmethod
    def calculate_platt(mol: Chem.Mol) -> float:
        """Get Platt number.
    
        Or Platt.
        """
        cc = [x.GetBeginAtom().GetDegree() + x.GetEndAtom().GetDegree() - 2 for x in mol.GetBonds()]
        return sum(cc)

    @staticmethod
    def calculate_simple_topo_index(mol: Chem.Mol) -> float:
        """Get the logarithm of the simple topological index.
    
        Or Sito.
        From Narumi H., MATCH (Comm. Math. Comp. Chem.), (1987), 22,195-207.
        """
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        while 0 in deltas:
            deltas.remove(0)
        deltas = np.array(deltas, 'd')
        res = np.prod(deltas)
        return np.log(res)

    @staticmethod
    def calculate_harmonic_topo_index(mol: Chem.Mol) -> float:
        """Get harmonic topological index.
    
        Or Hato.
        From Narumi H., MATCH (Comm. Math. Comp. Chem.), (1987), 22,195-207.
        """
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        while 0 in deltas:
            deltas.remove(0)
        if len(deltas) == 0:
            return 0.0
        deltas = np.array(deltas, 'd')
        nAtoms = mol.GetNumAtoms()
        res = nAtoms / sum(1. / deltas)
        return res

    @staticmethod
    def calculate_geometric_topo_index(mol: Chem.Mol) -> float:
        """Get Geometric topological index.
    
        Or Geto.
        From Narumi H., MATCH (Comm. Math. Comp. Chem.), (1987), 22,195-207.
        """
        nAtoms = mol.GetNumAtoms()
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        while 0 in deltas:
            deltas.remove(0)
        if len(deltas) == 0:
            return 0.0
        deltas = np.array(deltas, 'd')
        temp = np.prod(deltas)
        res = np.power(temp, 1. / nAtoms)
        return res

    @staticmethod
    def calculate_arithmetic_topo_index(mol: Chem.Mol) -> float:
        """Get Arithmetic topological index.
    
        Or Arto.
        From Narumi H., MATCH (Comm. Math. Comp. Chem.), (1987), 22,195-207.
        """
        nAtoms = mol.GetNumAtoms()
        nBonds = mol.GetNumBonds()
        res = 2. * nBonds / nAtoms
        return res

    @staticmethod
    def calculate_mol_size_total_inf(mol: Chem.Mol) -> float:
        """Get total information index on molecular size.
    
        Or ISIZ.
        """
        Hmol = Chem.AddHs(mol)
        nAT = Hmol.GetNumAtoms()
        ISIZ = nAT * np.log2(nAT)
        return ISIZ

    @staticmethod
    def calculate_atom_comp_total_inf(mol: Chem.Mol) -> float:
        """Ge total information index on atomic composition.
    
        Or TIAC.
        """
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = []
        for i in range(nAtoms):
            at = Hmol.GetAtomWithIdx(i)
            IC.append(at.GetAtomicNum())
        Unique = np.unique(IC)
        NAtomType = len(Unique)
        res = 0.0
        for i in range(NAtomType):
            cc = IC.count(Unique[i])
            res += cc * np.log2(cc)
        if nAtoms != 0:
            return nAtoms * np.log2(nAtoms) - res
        else:
            return 0.0

    @staticmethod
    def calculate_distance_equality_total_inf(mol: Chem.Mol) -> float:
        """Get total information index on distance equality.
    
        Or DET.
        """
        Distance = Chem.GetDistanceMatrix(mol)
        nAT = mol.GetNumAtoms()
        n = 1. / 2 * nAT ** 2 - nAT
        DisType = int(Distance.max())
        res = 0.0
        for i in range(DisType):
            cc = 1. / 2 * sum(sum(Distance == i + 1))
            res += cc * np.log2(cc)
        return n * np.log2(n) - res if n > 0 else np.NaN

    @staticmethod
    def _calculate_entropy(Probability: Iterable[float]) -> float:
        """calculate_ entropy (Information content) of given probability."""
        res = 0.0
        for i in Probability:
            if i != 0:
                res = res - i * np.log2(i)
        return res

    @staticmethod
    def calculate_distance_equality_mean_inf(mol: Chem.Mol) -> float:
        """Get the mean information index on distance equality.
    
        Or IDE.
        """
        Distance = Chem.GetDistanceMatrix(mol)
        nAT = mol.GetNumAtoms()
        n = 1. / 2 * nAT ** 2 - nAT
        DisType = int(Distance.max())
        res = 0.0
        cc = np.zeros(DisType, dtype=float)
        for i in range(DisType):
            cc[i] = 1. / 2 * sum(sum(Distance == i + 1))
        res = Topology._calculate_entropy(cc / n)
        return res

    @staticmethod
    def calculate_vertex_equality_total_inf(mol: Chem.Mol) -> float:
        """Get the total information index on vertex equality.
    
        Or IVDE.
        """
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        res = 0.0
        while 0 in deltas:
            deltas.remove()
        for i in range(max(deltas)):
            cc = deltas.count(i + 1)
            if cc == 0:
                res = res
            else:
                res += cc * np.log2(cc)
        n = len(deltas)
        return n * np.log2(n) - res

    @staticmethod
    def _hall_kier_deltas(mol: Chem.Mol, skipHs: bool = True) -> List[float]:
        """calculate_ Kier & Hall valence delta-values for molecular connectivity.
    
        From Kier L. and Hall L., J. Pharm. Sci. (1983), 72(10),1170-1173.
        """
        global periodicTable
        res = []
        for atom in mol.GetAtoms():
            n = atom.GetAtomicNum()
            if n > 1:
                nV = periodicTable.GetNOuterElecs(n)
                nHs = atom.GetTotalNumHs()
                if n < 10:
                    res.append(float(nV - nHs))
                else:
                    res.append(float(nV - nHs) / float(n - nV - 1))
            elif not skipHs:
                res.append(0.0)
        return res

    @staticmethod
    def calculate_simple_topo_vindex(mol: Chem.Mol) -> float:
        """Get the logarithm of the simple topological index.
    
        Or Sitov.
        From Narumi H., MATCH (Comm. Math. Comp. Chem.), (1987), 22,195-207.
    
        Kier and Hall's valence delta-values are used in place of atom degrees.
        From Kier L. and Hall L., J. Pharm. Sci. (1983), 72(10),1170-1173.
        """
        deltas = Topology._hall_kier_deltas(mol, skipHs=0)
        while 0 in deltas:
            deltas.remove(0)
        deltas = np.array(deltas, 'd')
        res = np.prod(deltas)
        return np.log(res)

    @staticmethod
    def calculate_harmonic_topo_vindex(mol: Chem.Mol) -> float:
        """Get harmonic topological index.
    
        Or Hatov.
        From Narumi H., MATCH (Comm. Math. Comp. Chem.), (1987), 22,195-207.
    
        Kier and Hall's valence delta-values are used in place of atom degrees.
        From Kier L. and Hall L., J. Pharm. Sci. (1983), 72(10),1170-1173.
        """
        deltas = Topology._hall_kier_deltas(mol, skipHs=0)
        while 0 in deltas:
            deltas.remove(0)
        deltas = np.array(deltas, 'd')
        nAtoms = mol.GetNumAtoms()
        res = nAtoms / sum(1. / deltas)
        return res

    @staticmethod
    def calculate_geometric_topo_vindex(mol: Chem.Mol) -> float:
        """Get Geometric topological index.
    
        Or Getov.
        From Narumi H., MATCH (Comm. Math. Comp. Chem.), (1987), 22,195-207.
    
        Kier and Hall's valence delta-values are used in place of atom degrees.
        From Kier L. and Hall L., J. Pharm. Sci. (1983), 72(10),1170-1173.
        """
        nAtoms = mol.GetNumAtoms()
        deltas = Topology._hall_kier_deltas(mol, skipHs=0)
        while 0 in deltas:
            deltas.remove(0)
        deltas = np.array(deltas, 'd')
        temp = np.prod(deltas)
        res = np.power(temp, 1. / nAtoms)
        return res

    @staticmethod
    def calculate_gravitational_topo_index(mol: Chem.Mol) -> float:
        """Get Gravitational topological index based on topological distance.
    
        Or Gravto
        From Katritzky, A. J. Phys. Chem., (1996), 100,10400-10407.
        """
        nAT = mol.GetNumAtoms()
        Distance = Chem.GetDistanceMatrix(mol)
        res = 0.0
        Atom = mol.GetAtoms()
        for i in range(nAT - 1):
            for j in range(i + 1, nAT):
                temp = Atom[i].GetMass() * Atom[j].GetMass()
                res = res + temp / np.power(Distance[i][j], 2)
        return res / 100

    @staticmethod
    def calculate_gutman_vtopo(mol: Chem.Mol) -> float:
        """Get molecular topological index based on valence vertex degree.
    
        Or GMTIV.
        From Gutman,I. J. Chem. Inf. Comput. Sci., (1994), 34,1037-1039.
        """
        mol_ = Chem.RemoveHs(mol)
        nAT = mol_.GetNumAtoms()
        deltas = Topology._hall_kier_deltas(mol_)
        Distance = Chem.GetDistanceMatrix(mol_)
        res = 0.0
        for i in range(nAT):
            for j in range(i + 1, nAT):
                res += deltas[i] * deltas[j] * Distance[i, j]
        return np.log10(res)
    
    
    _topology = {'W': calculate_weiner,
                 'AW': calculate_mean_weiner,
                 'J': calculate_balaban,
                 'Thara': calculate_harary,
                 'Tsch': calculate_schiultz,
                 'Tigdi': calculate_graph_distance,
                 'Platt': calculate_platt,
                 'Xu': calculate_xu_index,
                 'Pol': calculate_polarity_number,
                 'DZ': calculate_pogliani_index,
                 'Ipc': calculate_ipc,
                 'BertzCT': calculate_bertzct,
                 'GMTI': calculate_gutman_topo,
                 'ZM1': calculate_zagreb1,
                 'ZM2': calculate_zagreb2,
                 'MZM1': calculate_mzagreb1,
                 'MZM2': calculate_mzagreb2,
                 'Qindex': calculate_quadratic,
                 'diametert': calculate_diameter,
                 'radiust': calculate_radius,
                 'petitjeant': calculate_petitjean,
                 'Sito': calculate_simple_topo_index,
                 'Hato': calculate_harmonic_topo_index,
                 'Geto': calculate_geometric_topo_index,
                 'Arto': calculate_arithmetic_topo_index,
                 'ISIZ': calculate_mol_size_total_inf,
                 'TIAC': calculate_atom_comp_total_inf,
                 'IDET': calculate_distance_equality_total_inf,
                 'IDE': calculate_distance_equality_mean_inf,
                 'IVDE': calculate_vertex_equality_total_inf,
                 'Sitov': calculate_simple_topo_vindex,
                 'Hatov': calculate_harmonic_topo_vindex,
                 'Getov': calculate_geometric_topo_vindex,
                 'Gravto': calculate_gravitational_topo_index,
                 'GMTIV': calculate_gutman_vtopo,
                 }
    
    
    def get_all(mol: Chem.Mol) -> dict:
        """Get all (35) constitutional descriptors."""
        result = {}
        for DesLabel in Topology._topology.keys():
            result[DesLabel] = Topology._topology[DesLabel](mol)
        return result



# -*- coding: utf-8 -*-


"""Whole Holistic Invariant Molecular (WHIM) descriptors."""

from typing import List, Tuple

from rdkit import Chem
import numpy as np

from .atomproperty import get_relative_atomic_property
from .geo_opt import read_coordinates

class WHIM:
    """Whole Holistic Invariant Molecular descriptors."""

    @staticmethod
    def get_atom_coordinate_matrix(arc_file: str) -> Tuple[np.array, List[str]]:
        """Get atom coordinate matrix and elements.
    
        :param arc_file: Path to MOPAC .arc file
        :return: The 3D coordinates of the atoms of the optimized molecule
                 along with atomic symbols.
        """
        ChargeCoordinates = read_coordinates(arc_file)
        nAtom = len(ChargeCoordinates)
        CoordinateMatrix = np.zeros([nAtom, 3])
        AtomLabel = []
        for i, j in enumerate(ChargeCoordinates):
            CoordinateMatrix[i, :] = [j[1], j[2], j[3]]
            AtomLabel.append(j[0])
        return np.array(CoordinateMatrix), AtomLabel

    @staticmethod
    def x_pre_center(X: np.array) -> np.array:
        """Center the data matrix X."""
        Xdim = np.size(X, axis=0)
        Xmean = np.mean(X, axis=0)
        Xmean = np.array(Xmean)
        Xp = X - np.ones([Xdim, 1]) * Xmean
        return Xp

    @staticmethod
    def get_property_matrix(AtomLabel: List[str], proname: str = 'm') -> np.array:
        """Get the given property from an atomic symbol."""
        res = []
        for i in AtomLabel:
            res.append(get_relative_atomic_property(i, proname))
        return np.array(np.diag(res))

    @staticmethod
    def get_svd_eig(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> np.array:
        """Get singular values of the weighted covariance matrix."""
        nAtom, kc = CoordinateMatrix.shape
        if proname == 'u':
            weight = np.array(np.eye(nAtom))
        else:
            weight = WHIM.get_property_matrix(AtomLabel, proname)
        S = WHIM.x_pre_center(CoordinateMatrix)
        u, s, v = np.linalg.svd(S.T @ weight @ S / sum(np.diag(weight)))
        return s

    @staticmethod
    def get_whim1(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get L1u WHIM descriptor."""
        s = WHIM.get_svd_eig(CoordinateMatrix, AtomLabel, proname)
        return s[0]

    @staticmethod
    def get_whim2(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get L2u WHIM descriptor."""
        s = WHIM.get_svd_eig(CoordinateMatrix, AtomLabel, proname)
        return s[1]

    @staticmethod
    def get_whim3(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get L3u WHIM descriptor."""
        s = WHIM.get_svd_eig(CoordinateMatrix, AtomLabel, proname)
        return s[2]

    @staticmethod
    def get_whim4(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get Tu WHIM descriptor."""
        s = WHIM.get_svd_eig(CoordinateMatrix, AtomLabel, proname)
        T = sum(s)
        return T

    @staticmethod
    def get_whim5(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get Au WHIM descriptor."""
        s = WHIM.get_svd_eig(CoordinateMatrix, AtomLabel, proname)
        A = s[0] * s[1] + s[0] * s[2] + s[1] * s[2]
        return A

    @staticmethod
    def get_whim6(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get Vu WHIM descriptor."""
        s = WHIM.get_svd_eig(CoordinateMatrix, AtomLabel, proname)
        A = s[0] * s[1] + s[0] * s[2] + s[1] * s[2]
        T = sum(s)
        V = A + T + s[0] * s[1] * s[2]
        return V

    @staticmethod
    def get_whim7(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get P1u WHIM descriptor."""
        s = WHIM.get_svd_eig(CoordinateMatrix, AtomLabel, proname)
        return s[0] / (s[0] + s[1] + s[2])

    @staticmethod
    def get_whim8(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u'):
        """Get P2u WHIM descriptor."""
        s = WHIM.get_svd_eig(CoordinateMatrix, AtomLabel, proname)
        return s[1] / (s[0] + s[1] + s[2])

    @staticmethod
    def get_whim9(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get Ku WHIM descriptor."""
        s = WHIM.get_svd_eig(CoordinateMatrix, AtomLabel, proname)
        res = 0.0
        for i in s:
            res = res + abs(i / sum(s) - 1 / 3.0)
        Ku = 3.0 / 4 * res
        return Ku

    @staticmethod
    def get_whim10(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get E1u WHIM descriptor."""
        nAtom, kc = CoordinateMatrix.shape
        if proname == 'u':
            weight = np.array(np.eye(nAtom))
        else:
            weight = WHIM.get_property_matrix(AtomLabel, proname)
        S = WHIM.x_pre_center(CoordinateMatrix)
        u, s, v = np.linalg.svd(S.T @ weight @ S / sum(np.diag(weight)))
        res = np.power(s[0], 2) * nAtom / np.power(S * np.array(u[:, 0]).T, 4).sum()
        return float(res.real)

    @staticmethod
    def get_whim11(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get E2u WHIM descriptor."""
        nAtom, kc = CoordinateMatrix.shape
        if proname == 'u':
            weight = np.array(np.eye(nAtom))
        else:
            weight = WHIM.get_property_matrix(AtomLabel, proname)
        S = WHIM.x_pre_center(CoordinateMatrix)
        u, s, v = np.linalg.svd(S.T @ weight @ S / sum(np.diag(weight)))
        res = np.power(s[1], 2) * nAtom / np.power(S * np.array(u[:, 1]).T, 4).sum()
        return float(res.real)

    @staticmethod
    def get_whim12(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get E3u WHIM descriptor."""
        nAtom, kc = CoordinateMatrix.shape
        if proname == 'u':
            weight = np.array(np.eye(nAtom))
        else:
            weight = WHIM.get_property_matrix(AtomLabel, proname)
        S = WHIM.x_pre_center(CoordinateMatrix)
        u, s, v = np.linalg.svd(S.T @ weight @ S / sum(np.diag(weight)))
        res = np.power(s[2], 2) * nAtom / np.power(S * np.array(u[:, 2]).T, 4).sum()
        return float(res.real)

    @staticmethod
    def get_whim13(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get Du WHIM descriptor."""
        c1 = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname)
        c2 = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname)
        c3 = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname)
        Du = c1 + c2 + c3
        return float(Du)

    @staticmethod
    def get_whim14(CoordinateMatrix: np.array, AtomLabel: List[str], proname: str = 'u') -> float:
        """Get P3u WHIM descriptor."""
        s = WHIM.get_svd_eig(CoordinateMatrix, AtomLabel, proname)
        return s[2] / (s[0] + s[1] + s[2])

    @staticmethod
    def get_whim_unweighted() -> dict:
        """Get all unweighted WHIM descriptors."""
        res = {}
        CoordinateMatrix, AtomLabel = WHIM.get_atom_coordinate_matrix()
        res['L1u'] = WHIM.get_whim1(CoordinateMatrix, AtomLabel, proname='u')
        res['L2u'] = WHIM.get_whim2(CoordinateMatrix, AtomLabel, proname='u')
        res['L3u'] = WHIM.get_whim3(CoordinateMatrix, AtomLabel, proname='u')
        res['P1u'] = WHIM.get_whim7(CoordinateMatrix, AtomLabel, proname='u')
        res['P2u'] = WHIM.get_whim8(CoordinateMatrix, AtomLabel, proname='u')
        res['P3u'] = WHIM.get_whim14(CoordinateMatrix, AtomLabel, proname='u')
        res['E1u'] = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname='u')
        res['E2u'] = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname='u')
        res['E3u'] = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname='u')
        res['Tu'] = WHIM.get_whim4(CoordinateMatrix, AtomLabel, proname='u')
        res['Au'] = WHIM.get_whim5(CoordinateMatrix, AtomLabel, proname='u')
        res['Vu'] = WHIM.get_whim6(CoordinateMatrix, AtomLabel, proname='u')
        res['Ku'] = WHIM.get_whim9(CoordinateMatrix, AtomLabel, proname='u')
        res['Du'] = WHIM.get_whim13(CoordinateMatrix, AtomLabel, proname='u')
        return res

    @staticmethod
    def get_whim_mass() -> dict:
        """Get all WHIM descriptors based on atomic mass."""
        res = {}
        CoordinateMatrix, AtomLabel = WHIM.get_atom_coordinate_matrix()
        res['L1m'] = WHIM.get_whim1(CoordinateMatrix, AtomLabel, proname='m')
        res['L2m'] = WHIM.get_whim2(CoordinateMatrix, AtomLabel, proname='m')
        res['L3m'] = WHIM.get_whim3(CoordinateMatrix, AtomLabel, proname='m')
        res['Tm'] = WHIM.get_whim4(CoordinateMatrix, AtomLabel, proname='m')
        res['Am'] = WHIM.get_whim5(CoordinateMatrix, AtomLabel, proname='m')
        res['Vm'] = WHIM.get_whim6(CoordinateMatrix, AtomLabel, proname='m')
        res['P1m'] = WHIM.get_whim7(CoordinateMatrix, AtomLabel, proname='m')
        res['P2m'] = WHIM.get_whim8(CoordinateMatrix, AtomLabel, proname='m')
        res['Km'] = WHIM.get_whim9(CoordinateMatrix, AtomLabel, proname='m')
        res['E1m'] = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname='m')
        res['E2m'] = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname='m')
        res['E3m'] = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname='m')
        res['Dm'] = WHIM.get_whim13(CoordinateMatrix, AtomLabel, proname='m')
        res['P3m'] = WHIM.get_whim14(CoordinateMatrix, AtomLabel, proname='m')
        return res

    @staticmethod
    def get_whim_sanderson_electronegativity() -> dict:
        """Get all WHIM descriptors based on Sanderson electronegativity."""
        res = {}
        CoordinateMatrix, AtomLabel = WHIM.get_atom_coordinate_matrix()
        res['L1e'] = WHIM.get_whim1(CoordinateMatrix, AtomLabel, proname='En')
        res['L2e'] = WHIM.get_whim2(CoordinateMatrix, AtomLabel, proname='En')
        res['L3e'] = WHIM.get_whim3(CoordinateMatrix, AtomLabel, proname='En')
        res['P1e'] = WHIM.get_whim7(CoordinateMatrix, AtomLabel, proname='En')
        res['P2e'] = WHIM.get_whim8(CoordinateMatrix, AtomLabel, proname='En')
        res['P3e'] = WHIM.get_whim14(CoordinateMatrix, AtomLabel, proname='En')
        res['E1e'] = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname='En')
        res['E2e'] = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname='En')
        res['E3e'] = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname='En')
        res['Te'] = WHIM.get_whim4(CoordinateMatrix, AtomLabel, proname='En')
        res['Ae'] = WHIM.get_whim5(CoordinateMatrix, AtomLabel, proname='En')
        res['Ve'] = WHIM.get_whim6(CoordinateMatrix, AtomLabel, proname='En')
        res['Ke'] = WHIM.get_whim9(CoordinateMatrix, AtomLabel, proname='En')
        res['De'] = WHIM.get_whim13(CoordinateMatrix, AtomLabel, proname='En')
        return res

    @staticmethod
    def get_whim_vdw_volume() -> dict:
        """Get all WHIM descriptors based on vdW volume."""
        res = {}
        CoordinateMatrix, AtomLabel = WHIM.get_atom_coordinate_matrix()
        res['L1v'] = WHIM.get_whim1(CoordinateMatrix, AtomLabel, proname='V')
        res['L2v'] = WHIM.get_whim2(CoordinateMatrix, AtomLabel, proname='V')
        res['L3v'] = WHIM.get_whim3(CoordinateMatrix, AtomLabel, proname='V')
        res['P1v'] = WHIM.get_whim7(CoordinateMatrix, AtomLabel, proname='V')
        res['P2v'] = WHIM.get_whim8(CoordinateMatrix, AtomLabel, proname='V')
        res['P3v'] = WHIM.get_whim14(CoordinateMatrix, AtomLabel, proname='V')
        res['E1v'] = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname='V')
        res['E2v'] = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname='V')
        res['E3v'] = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname='V')
        res['Tv'] = WHIM.get_whim4(CoordinateMatrix, AtomLabel, proname='V')
        res['Av'] = WHIM.get_whim5(CoordinateMatrix, AtomLabel, proname='V')
        res['Vv'] = WHIM.get_whim6(CoordinateMatrix, AtomLabel, proname='V')
        res['Kv'] = WHIM.get_whim9(CoordinateMatrix, AtomLabel, proname='V')
        res['Dv'] = WHIM.get_whim13(CoordinateMatrix, AtomLabel, proname='V')
        return res

    @staticmethod
    def get_whim_polarizability() -> dict:
        """Get all WHIM descriptors based on polarizability."""
        res = {}
        CoordinateMatrix, AtomLabel = WHIM.get_atom_coordinate_matrix()
        res['L1p'] = WHIM.get_whim1(CoordinateMatrix, AtomLabel, proname='alapha')
        res['L2p'] = WHIM.get_whim2(CoordinateMatrix, AtomLabel, proname='alapha')
        res['L3p'] = WHIM.get_whim3(CoordinateMatrix, AtomLabel, proname='alapha')
        res['P1p'] = WHIM.get_whim7(CoordinateMatrix, AtomLabel, proname='alapha')
        res['P2p'] = WHIM.get_whim8(CoordinateMatrix, AtomLabel, proname='alapha')
        res['P3p'] = WHIM.get_whim14(CoordinateMatrix, AtomLabel, proname='alapha')
        res['E1p'] = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname='alapha')
        res['E2p'] = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname='alapha')
        res['E3p'] = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname='alapha')
        res['Tp'] = WHIM.get_whim4(CoordinateMatrix, AtomLabel, proname='alapha')
        res['Ap'] = WHIM.get_whim5(CoordinateMatrix, AtomLabel, proname='alapha')
        res['Vp'] = WHIM.get_whim6(CoordinateMatrix, AtomLabel, proname='alapha')
        res['Kp'] = WHIM.get_whim9(CoordinateMatrix, AtomLabel, proname='alapha')
        res['Dp'] = WHIM.get_whim13(CoordinateMatrix, AtomLabel, proname='alapha')
        return res

    @staticmethod
    def get_all(mol: Chem.Mol, arc_file: str) -> dict:
        """Get all (70) WHIM descriptors.
    
        :param dir_: Path to directory containing MOPAC .arc file
        """
        res = {}
        CoordinateMatrix, AtomLabel = WHIM.get_atom_coordinate_matrix(arc_file)
        res['L1u'] = WHIM.get_whim1(CoordinateMatrix, AtomLabel, proname='u')
        res['L2u'] = WHIM.get_whim2(CoordinateMatrix, AtomLabel, proname='u')
        res['L3u'] = WHIM.get_whim3(CoordinateMatrix, AtomLabel, proname='u')
        res['P1u'] = WHIM.get_whim7(CoordinateMatrix, AtomLabel, proname='u')
        res['P2u'] = WHIM.get_whim8(CoordinateMatrix, AtomLabel, proname='u')
        res['P3u'] = WHIM.get_whim14(CoordinateMatrix, AtomLabel, proname='u')
        res['E1u'] = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname='u')
        res['E2u'] = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname='u')
        res['E3u'] = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname='u')
        res['Tu'] = WHIM.get_whim4(CoordinateMatrix, AtomLabel, proname='u')
        res['Au'] = WHIM.get_whim5(CoordinateMatrix, AtomLabel, proname='u')
        res['Vu'] = WHIM.get_whim6(CoordinateMatrix, AtomLabel, proname='u')
        res['Ku'] = WHIM.get_whim9(CoordinateMatrix, AtomLabel, proname='u')
        res['Du'] = WHIM.get_whim13(CoordinateMatrix, AtomLabel, proname='u')
        res['L1m'] = WHIM.get_whim1(CoordinateMatrix, AtomLabel, proname='m')
        res['L2m'] = WHIM.get_whim2(CoordinateMatrix, AtomLabel, proname='m')
        res['L3m'] = WHIM.get_whim3(CoordinateMatrix, AtomLabel, proname='m')
        res['P1m'] = WHIM.get_whim7(CoordinateMatrix, AtomLabel, proname='m')
        res['P2m'] = WHIM.get_whim8(CoordinateMatrix, AtomLabel, proname='m')
        res['P3m'] = WHIM.get_whim14(CoordinateMatrix, AtomLabel, proname='m')
        res['E1m'] = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname='m')
        res['E2m'] = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname='m')
        res['E3m'] = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname='m')
        res['Tm'] = WHIM.get_whim4(CoordinateMatrix, AtomLabel, proname='m')
        res['Am'] = WHIM.get_whim5(CoordinateMatrix, AtomLabel, proname='m')
        res['Vm'] = WHIM.get_whim6(CoordinateMatrix, AtomLabel, proname='m')
        res['Km'] = WHIM.get_whim9(CoordinateMatrix, AtomLabel, proname='m')
        res['Dm'] = WHIM.get_whim13(CoordinateMatrix, AtomLabel, proname='m')
        res['L1e'] = WHIM.get_whim1(CoordinateMatrix, AtomLabel, proname='En')
        res['L2e'] = WHIM.get_whim2(CoordinateMatrix, AtomLabel, proname='En')
        res['L3e'] = WHIM.get_whim3(CoordinateMatrix, AtomLabel, proname='En')
        res['P1e'] = WHIM.get_whim7(CoordinateMatrix, AtomLabel, proname='En')
        res['P2e'] = WHIM.get_whim8(CoordinateMatrix, AtomLabel, proname='En')
        res['P3e'] = WHIM.get_whim14(CoordinateMatrix, AtomLabel, proname='En')
        res['E1e'] = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname='En')
        res['E2e'] = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname='En')
        res['E3e'] = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname='En')
        res['Te'] = WHIM.get_whim4(CoordinateMatrix, AtomLabel, proname='En')
        res['Ae'] = WHIM.get_whim5(CoordinateMatrix, AtomLabel, proname='En')
        res['Ve'] = WHIM.get_whim6(CoordinateMatrix, AtomLabel, proname='En')
        res['Ke'] = WHIM.get_whim9(CoordinateMatrix, AtomLabel, proname='En')
        res['De'] = WHIM.get_whim13(CoordinateMatrix, AtomLabel, proname='En')
        res['L1v'] = WHIM.get_whim1(CoordinateMatrix, AtomLabel, proname='V')
        res['L2v'] = WHIM.get_whim2(CoordinateMatrix, AtomLabel, proname='V')
        res['L3v'] = WHIM.get_whim3(CoordinateMatrix, AtomLabel, proname='V')
        res['P1v'] = WHIM.get_whim7(CoordinateMatrix, AtomLabel, proname='V')
        res['P2v'] = WHIM.get_whim8(CoordinateMatrix, AtomLabel, proname='V')
        res['P3v'] = WHIM.get_whim14(CoordinateMatrix, AtomLabel, proname='V')
        res['E1v'] = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname='V')
        res['E2v'] = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname='V')
        res['E3v'] = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname='V')
        res['Tv'] = WHIM.get_whim4(CoordinateMatrix, AtomLabel, proname='V')
        res['Av'] = WHIM.get_whim5(CoordinateMatrix, AtomLabel, proname='V')
        res['Vv'] = WHIM.get_whim6(CoordinateMatrix, AtomLabel, proname='V')
        res['Kv'] = WHIM.get_whim9(CoordinateMatrix, AtomLabel, proname='V')
        res['Dv'] = WHIM.get_whim13(CoordinateMatrix, AtomLabel, proname='V')
        res['L1p'] = WHIM.get_whim1(CoordinateMatrix, AtomLabel, proname='alapha')
        res['L2p'] = WHIM.get_whim2(CoordinateMatrix, AtomLabel, proname='alapha')
        res['L3p'] = WHIM.get_whim3(CoordinateMatrix, AtomLabel, proname='alapha')
        res['P1p'] = WHIM.get_whim7(CoordinateMatrix, AtomLabel, proname='alapha')
        res['P2p'] = WHIM.get_whim8(CoordinateMatrix, AtomLabel, proname='alapha')
        res['P3p'] = WHIM.get_whim14(CoordinateMatrix, AtomLabel, proname='alapha')
        res['E1p'] = WHIM.get_whim10(CoordinateMatrix, AtomLabel, proname='alapha')
        res['E2p'] = WHIM.get_whim11(CoordinateMatrix, AtomLabel, proname='alapha')
        res['E3p'] = WHIM.get_whim12(CoordinateMatrix, AtomLabel, proname='alapha')
        res['Tp'] = WHIM.get_whim4(CoordinateMatrix, AtomLabel, proname='alapha')
        res['Ap'] = WHIM.get_whim5(CoordinateMatrix, AtomLabel, proname='alapha')
        res['Vp'] = WHIM.get_whim6(CoordinateMatrix, AtomLabel, proname='alapha')
        res['Kp'] = WHIM.get_whim9(CoordinateMatrix, AtomLabel, proname='alapha')
        res['Dp'] = WHIM.get_whim13(CoordinateMatrix, AtomLabel, proname='alapha')
        return res



# -*- coding: utf-8 -*-


"""Constant for chemopy unittesting."""

import os

from rdkit import Chem
from rdkit.Chem import AllChem, rdDistGeom


SAMPLES_FOLDER = os.path.realpath(os.path.join(os.path.dirname(__file__), 'samples'))

MOL_MOLECULE = Chem.MolFromMolFile(os.path.join(SAMPLES_FOLDER, 'CID148124.mol'))
MOL2_MOLECULE = Chem.MolFromMol2File(os.path.join(SAMPLES_FOLDER, 'CID148124.mol2'))

with open(os.path.join(SAMPLES_FOLDER, 'testmol.smi')) as fh:
    SMILES_MOLECULES = [Chem.MolFromSmiles(line.strip()) for line in fh]

with open(os.path.join(SAMPLES_FOLDER, 'cas.txt')) as fh:
    CAS_NUMBERS = [line.strip() for line in fh.readlines()]

with open(os.path.join(SAMPLES_FOLDER, 'drugbank.txt')) as fh:
    DRUGBANK_NUMBERS = [line.strip() for line in fh.readlines()]

with open(os.path.join(SAMPLES_FOLDER, 'kegg.txt')) as fh:
    KEGG_NUMBERS = [line.strip() for line in fh.readlines()]

with open(os.path.join(SAMPLES_FOLDER, 'ncbi.txt')) as fh:
    NCBI_NUMBERS = [line.strip() for line in fh.readlines()]

with open(os.path.join(SAMPLES_FOLDER, 'ebi.txt')) as fh:
    EBI_NUMBERS = [line.strip() for line in fh.readlines()]

MOLECULES_2D = [MOL_MOLECULE, MOL2_MOLECULE] + SMILES_MOLECULES

MOLECULES_3D = [Chem.AddHs(mol) for mol in MOLECULES_2D]
_embedded_3d = map(AllChem.EmbedMolecule, MOLECULES_3D)
MOLECULES_3D = [mol for mol, success in zip(MOLECULES_3D, _embedded_3d) if success == 0]



# -*- coding: utf-8 -*-

"""Molecular connectivity topological indices."""

from typing import List

import numpy as np
from rdkit import Chem
from rdkit.Chem import rdchem

from .topology import Topology

periodicTable = rdchem.GetPeriodicTable()


class Connectivity:
    """Topological indices"""

    @staticmethod
    def calculate_chi0(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 0."""
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        while [0] in deltas:
            deltas.remove([0])
        deltas = np.array(deltas, 'd')
        res = sum(np.sqrt(1. / deltas))
        return res

    @staticmethod
    def calculate_chi1(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 1."""
        cc = [x.GetBeginAtom().GetDegree() * x.GetEndAtom().GetDegree() for x in mol.GetBonds()]
        if len(cc) == 0:
            return 0.0
        while 0 in cc:
            cc.remove(0)
        cc = np.array(cc, 'd')
        res = sum(np.sqrt(1. / cc))
        return res

    @staticmethod
    def calculate_mean_randic(mol: Chem.Mol) -> float:
        """Calculate mean chi1 (Randic) connectivity index."""
        cc = [x.GetBeginAtom().GetDegree() * x.GetEndAtom().GetDegree() for x in mol.GetBonds()]
        if len(cc) == 0:
            return 0.0
        while 0 in cc:
            cc.remove(0)
        cc = np.array(cc, 'd')
        res = np.mean(np.sqrt(1. / cc))
        return res

    @staticmethod
    def _calculate_chinp(mol: Chem.Mol, NumPath: int = 2) -> float:
        """Calculate molecular connectivity chi index for path order 2."""
        accum = 0.0
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        for path in Chem.FindAllPathsOfLengthN(mol, NumPath + 1, useBonds=0):
            cAccum = 1.0
            for idx in path:
                cAccum *= deltas[idx]
            if cAccum:
                accum += 1. / np.sqrt(cAccum)
        return accum

    @staticmethod
    def calculate_chi2(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 2."""
        return Connectivity._calculate_chinp(mol, NumPath=2)

    @staticmethod
    def calculate_chi3p(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 3."""
        return Connectivity._calculate_chinp(mol, NumPath=3)

    @staticmethod
    def calculate_chi4p(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 4."""
        return Connectivity._calculate_chinp(mol, NumPath=4)

    @staticmethod
    def calculate_chi5p(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 5."""
        return Connectivity._calculate_chinp(mol, NumPath=5)

    @staticmethod
    def calculate_chi6p(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 6."""
        return Connectivity._calculate_chinp(mol, NumPath=6)

    @staticmethod
    def calculate_chi7p(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 7."""
        return Connectivity._calculate_chinp(mol, NumPath=7)

    @staticmethod
    def calculate_chi8p(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 8."""
        return Connectivity._calculate_chinp(mol, NumPath=8)

    @staticmethod
    def calculate_chi9p(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 9."""
        return Connectivity._calculate_chinp(mol, NumPath=9)

    @staticmethod
    def calculate_chi10p(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path order 10."""
        return Connectivity._calculate_chinp(mol, NumPath=10)

    @staticmethod
    def calculate_chi3c(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for cluster."""
        accum = 0.0
        patt = Chem.MolFromSmarts('*~*(~*)~*')
        HPatt = mol.GetSubstructMatches(patt)
        for cluster in HPatt:
            deltas = [mol.GetAtomWithIdx(x).GetDegree() for x in cluster]
            while [0] in deltas:
                deltas.remove([0])
            if deltas != []:
                deltas1 = np.array(deltas, dtype=float)
                accum = accum + 1. / np.sqrt(deltas1.prod())
        return accum

    @staticmethod
    def calculate_chi4c(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for cluster."""
        accum = 0.0
        patt = Chem.MolFromSmarts('*~*(~*)(~*)~*')
        HPatt = mol.GetSubstructMatches(patt)
        for cluster in HPatt:
            deltas = [mol.GetAtomWithIdx(x).GetDegree() for x in cluster]
            while [0] in deltas:
                deltas.remove([0])
            if deltas != []:
                deltas1 = np.array(deltas, dtype=float)
                accum = accum + 1. / np.sqrt(deltas1.prod())
        return accum

    @staticmethod
    def calculate_chi4pc(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for path/cluster."""
        accum = 0.0
        patt = Chem.MolFromSmarts('*~*(~*)~*~*')
        HPatt = mol.GetSubstructMatches(patt)
        for cluster in HPatt:
            deltas = [mol.GetAtomWithIdx(x).GetDegree() for x in cluster]
            while [0] in deltas:
                deltas.remove([0])
            if deltas != []:
                deltas1 = np.array(deltas, dtype=float)
                accum = accum + 1. / np.sqrt(deltas1.prod())
        return accum

    @staticmethod
    def calculate_delta_chi3c4pc(mol: Chem.Mol) -> float:
        """Calculate the difference between chi3c and chi4pc."""
        return abs(Connectivity.calculate_chi3c(mol) - Connectivity.calculate_chi4pc(mol))

    @staticmethod
    def _calculate_chinch(mol: Chem.Mol, NumCycle=3) -> float:
        """Calculate molecular connectivity chi index for cycles of n."""
        accum = 0.0
        deltas = [x.GetDegree() for x in mol.GetAtoms()]
        for tup in mol.GetRingInfo().AtomRings():
            cAccum = 1.0
            if len(tup) == NumCycle:
                for idx in tup:
                    cAccum *= deltas[idx]
                if cAccum:
                    accum += 1. / np.sqrt(cAccum)
        return accum

    @staticmethod
    def calculate_chi3ch(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for cycles of 3."""
        return Connectivity._calculate_chinch(mol, NumCycle=3)

    @staticmethod
    def calculate_chi4ch(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for cycles of 4."""
        return Connectivity._calculate_chinch(mol, NumCycle=4)

    @staticmethod
    def calculate_chi5ch(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for cycles of 5."""
        return Connectivity._calculate_chinch(mol, NumCycle=5)

    @staticmethod
    def calculate_chi6ch(mol: Chem.Mol) -> float:
        """Calculate molecular connectivity chi index for cycles of 6."""
        return Connectivity._calculate_chinch(mol, NumCycle=6)

    @staticmethod
    def calculate_chiv0(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 0."""
        deltas = Topology._hall_kier_deltas(mol, skipHs=0)
        while 0 in deltas:
            deltas.remove(0)
        deltas = np.array(deltas, 'd')
        res = sum(np.sqrt(1. / deltas))
        return res

    @staticmethod
    def _calculate_chivnp(mol: Chem.Mol, NumPath: int = 1) -> float:
        """Calculate valence molecular connectivity chi index for path order 1."""
        accum = 0.0
        deltas = Topology._hall_kier_deltas(mol, skipHs=False)
        for path in Chem.FindAllPathsOfLengthN(mol, NumPath + 1, useBonds=0):
            cAccum = 1.0
            for idx in path:
                cAccum *= deltas[idx]
            if cAccum:
                accum += 1. / np.sqrt(cAccum)
        return accum

    @staticmethod
    def calculate_chiv1(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 1."""
        return Connectivity._calculate_chivnp(mol, NumPath=1)

    @staticmethod
    def calculate_chiv2(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 2."""
        return Connectivity._calculate_chivnp(mol, NumPath=2)

    @staticmethod
    def calculate_chiv3p(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 3."""
        return Connectivity._calculate_chivnp(mol, NumPath=3)

    @staticmethod
    def calculate_chiv4p(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 4."""
        return Connectivity._calculate_chivnp(mol, NumPath=4)

    @staticmethod
    def calculate_chiv5p(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 5."""
        return Connectivity._calculate_chivnp(mol, NumPath=5)

    @staticmethod
    def calculate_chiv6p(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 6."""
        return Connectivity._calculate_chivnp(mol, NumPath=6)

    @staticmethod
    def calculate_chiv7p(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 7."""
        return Connectivity._calculate_chivnp(mol, NumPath=7)

    @staticmethod
    def calculate_chiv8p(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 8."""
        return Connectivity._calculate_chivnp(mol, NumPath=8)

    @staticmethod
    def calculate_chiv9p(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 9."""
        return Connectivity._calculate_chivnp(mol, NumPath=9)

    @staticmethod
    def calculate_chiv10p(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path order 10."""
        return Connectivity._calculate_chivnp(mol, NumPath=10)

    @staticmethod
    def calculate_delta_chi0(mol: Chem.Mol) -> float:
        """Calculate the difference between chi0v and chi0."""
        return abs(Connectivity.calculate_chiv0(mol) - Connectivity.calculate_chi0(mol))

    @staticmethod
    def calculate_delta_chi1(mol: Chem.Mol) -> float:
        """Calculate the difference between chi1v and chi1."""
        return abs(Connectivity.calculate_chiv1(mol) - Connectivity.calculate_chi1(mol))

    @staticmethod
    def calculate_delta_chi2(mol: Chem.Mol) -> float:
        """Calculate the difference between chi2v and chi2."""
        return abs(Connectivity._calculate_chivnp(mol, NumPath=2) - Connectivity._calculate_chinp(mol, NumPath=2))

    @staticmethod
    def calculate_delta_chi3(mol: Chem.Mol) -> float:
        """Calculate the difference between chi3v and chi3."""
        return abs(Connectivity._calculate_chivnp(mol, NumPath=3) - Connectivity._calculate_chinp(mol, NumPath=3))

    @staticmethod
    def calculate_delta_chi4(mol: Chem.Mol) -> float:
        """Calculate the difference between chi4v and chi4."""
        return abs(Connectivity._calculate_chivnp(mol, NumPath=4) - Connectivity._calculate_chinp(mol, NumPath=4))

    @staticmethod
    def _atom_hall_kier_deltas(atom: Chem.Atom, skipHs: bool = False) -> List[float]:
        """Calculate Kier & Hall atomic valence delta-values for molecular connectivity.

        From Kier L. and Hall L., J. Pharm. Sci. (1983), 72(10),1170-1173.
        """
        global periodicTable
        res = []
        n = atom.GetAtomicNum()
        if n > 1:
            nV = periodicTable.GetNOuterElecs(n)
            nHs = atom.GetTotalNumHs()
            if n < 10:
                res.append(float(nV - nHs))
            else:
                res.append(float(nV - nHs) / float(n - nV - 1))
        elif not skipHs:
            res.append(0.0)
        return res

    @staticmethod
    def calculate_chiv3c(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for cluster."""
        accum = 0.0
        patt = Chem.MolFromSmarts('*~*(~*)~*')
        HPatt = mol.GetSubstructMatches(patt)
        for cluster in HPatt:
            deltas = [Connectivity._atom_hall_kier_deltas(mol.GetAtomWithIdx(x)) for x in cluster]
            while [0] in deltas:
                deltas.remove([0])
            if deltas != []:
                deltas1 = np.array(deltas, dtype=float)
                accum = accum + 1. / np.sqrt(deltas1.prod())
        return accum

    @staticmethod
    def calculate_chiv4c(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for cluster."""
        accum = 0.0
        patt = Chem.MolFromSmarts('*~*(~*)(~*)~*')
        HPatt = mol.GetSubstructMatches(patt)
        for cluster in HPatt:
            deltas = [Connectivity._atom_hall_kier_deltas(mol.GetAtomWithIdx(x)) for x in cluster]
            while [0] in deltas:
                deltas.remove([0])
            if deltas != []:
                deltas1 = np.array(deltas, dtype=float)
                accum = accum + 1. / np.sqrt(deltas1.prod())
        return accum

    @staticmethod
    def calculate_chiv4pc(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for path/cluster."""
        accum = 0.0
        patt = Chem.MolFromSmarts('*~*(~*)~*~*')
        HPatt = mol.GetSubstructMatches(patt)
        for cluster in HPatt:
            deltas = [Connectivity._atom_hall_kier_deltas(mol.GetAtomWithIdx(x)) for x in cluster]
            while [0] in deltas:
                deltas.remove([0])
            if deltas != []:
                deltas1 = np.array(deltas, dtype=float)
                accum = accum + 1. / np.sqrt(deltas1.prod())
        return accum

    @staticmethod
    def calculate_delta_chiv3c4pc(mol: Chem.Mol) -> float:
        """Calculate the difference between chiv3c and chiv4pc."""
        return abs(Connectivity.calculate_chiv3c(mol) - Connectivity.calculate_chiv4pc(mol))

    @staticmethod
    def _calculate_chivnch(mol: Chem.Mol, NumCyc=3) -> float:
        """Calculate valence molecular connectivity chi index for cycles of n."""
        accum = 0.0
        deltas = Topology._hall_kier_deltas(mol, skipHs=0)
        for tup in mol.GetRingInfo().AtomRings():
            cAccum = 1.0
            if len(tup) == NumCyc:
                for idx in tup:
                    cAccum *= deltas[idx]
                if cAccum:
                    accum += 1. / np.sqrt(cAccum)
        return accum

    @staticmethod
    def calculate_chiv3ch(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for cycles of 3."""
        return Connectivity._calculate_chivnch(mol, NumCyc=3)

    @staticmethod
    def calculate_chiv4ch(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for cycles of 4."""
        return Connectivity._calculate_chivnch(mol, NumCyc=4)

    @staticmethod
    def calculate_chiv5ch(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for cycles of 5."""
        return Connectivity._calculate_chivnch(mol, NumCyc=5)

    @staticmethod
    def calculate_chiv6ch(mol: Chem.Mol) -> float:
        """Calculate valence molecular connectivity chi index for cycles of 6."""
        return Connectivity._calculate_chivnch(mol, NumCyc=6)


    _connectivity = {'Chiv0': calculate_chiv0,
                     'Chiv1': calculate_chiv1,
                     'Chiv2': calculate_chiv2,
                     'Chiv3': calculate_chiv3p,
                     'Chiv4': calculate_chiv4p,
                     'Chiv5': calculate_chiv5p,
                     'Chiv6': calculate_chiv6p,
                     'Chiv7': calculate_chiv7p,
                     'Chiv8': calculate_chiv8p,
                     'Chiv9': calculate_chiv9p,
                     'Chiv10': calculate_chiv10p,
                     'Chi3c': calculate_chi3c,
                     'Chi4c': calculate_chi4c,
                     'Chi4pc': calculate_chi4pc,
                     'Chi3ch': calculate_chi3ch,
                     'Chi4ch': calculate_chi4ch,
                     'Chi5ch': calculate_chi5ch,
                     'Chi6ch': calculate_chi6ch,
                     'Chi0': calculate_chi0,
                     'Chi1': calculate_chi1,
                     'Chi2': calculate_chi2,
                     'Chi3': calculate_chi3p,
                     'Chi4': calculate_chi4p,
                     'Chi5': calculate_chi5p,
                     'Chi6': calculate_chi6p,
                     'Chi7': calculate_chi7p,
                     'Chi8': calculate_chi8p,
                     'Chi9': calculate_chi9p,
                     'Chi10': calculate_chi10p,
                     'Chiv3c': calculate_chiv3c,
                     'Chiv4c': calculate_chiv4c,
                     'Chiv4pc': calculate_chiv4pc,
                     'Chiv3ch': calculate_chiv3ch,
                     'Chiv4ch': calculate_chiv4ch,
                     'Chiv5ch': calculate_chiv5ch,
                     'Chiv6ch': calculate_chiv6ch,
                     'mChi1': calculate_mean_randic,
                     'knotp': calculate_delta_chi3c4pc,
                     'dchi0': calculate_delta_chi0,
                     'dchi1': calculate_delta_chi1,
                     'dchi2': calculate_delta_chi2,
                     'dchi3': calculate_delta_chi3,
                     'dchi4': calculate_delta_chi4,
                     'knotpv': calculate_delta_chiv3c4pc}

    @staticmethod
    def get_all(mol: Chem.Mol) -> dict:
        """Get all (44) connectivity descriptors."""
        result = {}
        for DesLabel in Connectivity._connectivity.keys():
            result[DesLabel] = Connectivity._connectivity[DesLabel](mol)
        return result



# -*- coding: utf-8 -*-


"""A package to calculate molecular descriptors."""


from .chemopy import ChemoPy  # noqa: F401
from .getmol import MolFrom   # noqa: F401
from .fingerprint import Fingerprint # noqa: F401
from .fingerprint3d import Fingerprint3D # noqa: F401


__version__ = '1.0.1'



# -*- coding: utf-8 -*-


"""MOE-type descriptors.

Includes LabuteASA, TPSA, slogPVSA, MRVSA,
         PEOEVSA, EstateVSA and VSAEstate.
"""

from typing import List

from rdkit import Chem
from rdkit.Chem import MolSurf
from rdkit.Chem.EState import EState_VSA as EVSA


class MOE:
    """MOE-type 2D descriptors."""

    @staticmethod
    def calculate_labute_asa(mol: Chem.Mol) -> dict:
        """Calculate Labute's Approximate Surface Area (ASA from MOE)."""
        asa = MolSurf.pyLabuteASA(mol, includeHs=1)
        return asa

    @staticmethod
    def calculate_tpsa(mol: Chem.Mol) -> dict:
        """Calculate topological polar surface area based on fragments.

        Implementation based on the Daylight contrib program TPSA.
        """
        tpsa = MolSurf.TPSA(mol)
        return tpsa

    @staticmethod
    def calculate_slogpvsa(mol: Chem.Mol, bins: List[float] = None) -> dict:
        """Get MOE-type descriptors using LogP and SA contributions.

        :param bins: interval boundaries used in the P_VSA calculation.
                     The default SLOGP bins are [-0.4,-0.2,0,0.1,0.15,0.2,0.25,0.3,0.4,0.5,0.6].
        """
        temp = MolSurf.SlogP_VSA_(mol, bins, force=1)
        res = {}
        for i, j in enumerate(temp):
            res[f'slogPVSA{i}'] = j
        return res

    @staticmethod
    def calculate_smrvsa(mol: Chem.Mol, bins: List[float] = None) -> dict:
        """Get MOE-type descriptors using MR and SA contributions.

        :param bins: interval boundaries used in the P_VSA calculation.
                     The default SMR bins are [1.29, 1.82, 2.24, 2.45, 2.75, 3.05, 3.63,3.8,4.0].
        """
        temp = MolSurf.SMR_VSA_(mol, bins, force=1)
        res = {}
        for i, j in enumerate(temp):
            res[f'SMRVSA{i}'] = j
        return res

    @staticmethod
    def calculate_peoevsa(mol: Chem.Mol, bins: List[float] = None) -> dict:
        """Get MOE-type descriptors using partial charges and SA contributions.

        :param bins: interval boundaries used in the P_VSA calculation.
                     The default PEOE bins are [-.3,-.25,-.20,-.15,-.10,-.05,0,.05,.10,.15,.20,.25,.30].
        """
        temp = MolSurf.PEOE_VSA_(mol, bins, force=1)
        res = {}
        for i, j in enumerate(temp):
            res[f'PEOEVSA{i}'] = j
        return res

    @staticmethod
    def calculate_estate_vsa(mol: Chem.Mol, bins: List[float] = None) -> dict:
        """Get MOE-type descriptors using Estate indices and SA contributions.

        :param bins: interval boundaries used in the P_VSA calculation.
                     The default Estate bins are [-0.390,0.290,0.717,1.165,1.540,1.807,2.05,4.69,9.17,15.0].
        """
        temp = EVSA.EState_VSA_(mol, bins, force=1)
        res = {}
        for i, j in enumerate(temp):
            res[f'EstateVSA{i}'] = j
        return res

    @staticmethod
    def calculate_vsa_estate(mol: Chem.Mol, bins: List[float] = None) -> dict:
        """Get MOE-type descriptors using SA and Estate indices contributions.

        :param bins: interval boundaries used in the P_VSA calculation.
                     The default VSA bins are [4.78,5.00,5.410,5.740,6.00,6.07,6.45,7.00,11.0].
        """
        temp = EVSA.VSA_EState_(mol, bins, force=1)
        res = {}
        for i, j in enumerate(temp):
            res[f'VSAEstate{i}'] = j
        return res

    @staticmethod
    def get_all(mol: Chem.Mol) -> dict:
        """Calculate all (59) MOE-type descriptors."""
        result = {}
        result['TPSA1'] = MOE.calculate_tpsa(mol)
        result['LabuteASA'] = MOE.calculate_labute_asa(mol)
        result.update(MOE.calculate_slogpvsa(mol, bins=None))  # 12 values
        result.update(MOE.calculate_smrvsa(mol, bins=None))  # 10 values
        result.update(MOE.calculate_peoevsa(mol, bins=None))  # 14 values
        result.update(MOE.calculate_estate_vsa(mol, bins=None))  # 11 values
        result.update(MOE.calculate_vsa_estate(mol, bins=None))  # 10 values
        return result



# -*- coding: utf-8 -*-


"""Molecular physical/chemical properties."""


import math
import os
import subprocess  # noqa: S404
import tempfile
from platform import architecture
from sys import platform
from typing import List, Union

from openbabel import pybel
from rdkit import Chem
from rdkit.Chem import Crippen
from rdkit.Chem import MolSurf

from .geo_opt import clean_mopac_files


class MolecularProperties:
    """Molecular physical/chemical properties."""

    @staticmethod
    def calculate_mollogp(mol: Chem.Mol) -> float:
        """Cacluate of MolLogP.

        From Wildman and G. M. Crippen JCICS _39_ 868-873 (1999).
        """
        return Crippen._pyMolLogP(mol)

    @staticmethod
    def calculate_mollogp2(mol: Chem.Mol) -> float:
        """Cacluate MolLogP^2.

        From Wildman and G. M. Crippen JCICS _39_ 868-873 (1999).
        """
        res = Crippen._pyMolLogP(mol)
        return res * res

    @staticmethod
    def calculate_molmr(mol: Chem.Mol) -> float:
        """Cacluate molecular refraction.

        From Wildman and G. M. Crippen JCICS _39_ 868-873 (1999).
        """
        return Crippen._pyMolMR(mol)

    @staticmethod
    def calculate_tpsa(mol: Chem.Mol) -> float:
        """Calculate the topological polar surface area.

        From Ertl P. et al., J.Med.Chem. (2000), 43,3714-3717.
        """
        return MolSurf.TPSA(mol)

    @staticmethod
    def _calculate_bond_number(mol: Chem.Mol, bondtype: str = 'SINGLE') -> float:
        """Calculate number of bond of specified type.

        :param bondtype: can be SINGLE, DOUBLE, TRIPLE or AROMATIC.
        """
        i = 0
        for bond in mol.GetBonds():
            if bond.GetBondType().name == bondtype:
                i += 1
        return i

    @staticmethod
    def calculate_unsaturation_index(mol: Chem.Mol) -> float:
        """Calculate unsaturation index."""
        nd = MolecularProperties._calculate_bond_number(mol, bondtype='DOUBLE')
        nt = MolecularProperties._calculate_bond_number(mol, bondtype='TRIPLE')
        na = MolecularProperties._calculate_bond_number(mol, bondtype='AROMATIC')
        res = math.log((1 + nd + nt + na), 2)
        return res

    @staticmethod
    def calculate_hydrophilicity_factor(mol: Chem.Mol) -> float:
        """Calculate hydrophilicity factor.

        From Todeschini R. et al., SAR QSAR Environ Res (1997), 7,173-193.
        """
        nheavy = mol.GetNumHeavyAtoms()
        nc = 0
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() == 6:
                nc += 1
        nhy = 0
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() == 7 or atom.GetAtomicNum() == 8 or atom.GetAtomicNum() == 16:
                atomn = atom.GetNeighbors()
                for i in atomn:
                    if i.GetAtomicNum() == 1:
                        nhy += 1
        res = ((1 + nhy) * math.log((1 + nhy), 2) + nc * (1.0 / nheavy * math.log(1.0 / nheavy, 2)) + math.sqrt(
            (nhy + 0.0) / (nheavy * nheavy))) / math.log(1.0 + nheavy)
        return res


    molecular_property = {'MR': calculate_molmr,
                          'LogP': calculate_mollogp,
                          'LogP2': calculate_mollogp2,
                          'TPSA': calculate_tpsa,
                          'UI': calculate_unsaturation_index,
                          'Hy': calculate_hydrophilicity_factor,
                          }

    @staticmethod
    def get_all(mol: Chem.Mol) -> dict:
        """Get all (6) constitutional descriptors."""
        result = {}
        for des_label in MolecularProperties.molecular_property.keys():
            result[des_label] = MolecularProperties.molecular_property[des_label](mol)
        return result



# -*- coding: utf-8 -*-


"""Estate fingerprints and values.

Based on Hall, Money, Kier, J. Chem. Inf. Comput. Sci. (1991)
doi: 10.1021/ci00001a012
"""


import numpy as np
from rdkit import Chem
from rdkit.Chem.EState import Fingerprinter as ESFP

from . import atomtypes as ATEstate

class EState:
    """Electrotopological descriptors and fingerprints."""

    @staticmethod
    def _calculate_estate(mol: Chem.Mol, skipH: bool = True) -> np.array:
        """Get the EState value of each atom in the molecule."""
        mol = Chem.AddHs(mol)
        if skipH:
            mol = Chem.RemoveHs(mol)
        tb1 = Chem.GetPeriodicTable()
        nAtoms = mol.GetNumAtoms()
        Is = np.zeros(nAtoms, dtype=float)
        for i in range(nAtoms):
            at = mol.GetAtomWithIdx(i)
            atNum = at.GetAtomicNum()
            d = at.GetDegree()
            if d > 0:
                h = at.GetTotalNumHs()
                dv = tb1.GetNOuterElecs(atNum) - h
                N = EState._get_principle_quantum_number(atNum)
                Is[i] = (4.0 / (N * N) * dv + 1) / d
        dists = Chem.GetDistanceMatrix(mol, useBO=0, useAtomWts=0)
        dists += 1
        accum = np.zeros(nAtoms, dtype=float)
        for i in range(nAtoms):
            for j in range(i + 1, nAtoms):
                p = dists[i, j]
                if p < 1e6:
                    temp = (Is[i] - Is[j]) / (p * p)
                    accum[i] += temp
                    accum[j] -= temp
        res = accum + Is
        return res

    @staticmethod
    def _get_principle_quantum_number(atNum: int) -> int:
        """Get the principle quantum number of atom from atomic number."""
        if atNum <= 2:
            return 1
        elif atNum <= 10:
            return 2
        elif atNum <= 18:
            return 3
        elif atNum <= 36:
            return 4
        elif atNum <= 54:
            return 5
        elif atNum <= 86:
            return 6
        else:
            return 7

    @staticmethod
    def calculate_heavy_atom_estate(mol: Chem.Mol) -> float:
        """Calculate sum of the EState indices over all heavy atoms."""
        return sum(EState._calculate_estate(mol))

    @staticmethod
    def _calculate_atom_estate(mol: Chem.Mol, atomic_num=6) -> float:
        """Calculate the sum of the EState indices over all atoms with specified atomic number."""
        nAtoms = mol.GetNumAtoms()
        Is = np.zeros(nAtoms, dtype=float)
        Estate = EState._calculate_estate(mol, skipH=False)
        for i in range(nAtoms):
            at = mol.GetAtomWithIdx(i)
            atNum = at.GetAtomicNum()
            if atNum == atomic_num:
                Is[i] = Estate[i]
        res = sum(Is)
        return res

    @staticmethod
    def calculate_c_atom_estate(mol: Chem.Mol) -> float:
        """Calculate the sum of the EState indices over all C atoms."""
        return EState._calculate_atom_estate(mol, atomic_num=6)

    @staticmethod
    def calculate_halogen_estate(mol: Chem.Mol) -> float:
        """Calculate the sum of the EState indices over all Halogen atoms."""
        Nf = EState._calculate_atom_estate(mol, atomic_num=9)
        Ncl = EState._calculate_atom_estate(mol, atomic_num=17)
        Nbr = EState._calculate_atom_estate(mol, atomic_num=35)
        Ni = EState._calculate_atom_estate(mol, atomic_num=53)
        return Nf + Ncl + Nbr + Ni

    @staticmethod
    def calculate_hetero_estate(mol: Chem.Mol) -> float:
        """Calculate the sum of the EState indices over all hetero atoms."""
        Ntotal = sum(EState._calculate_estate(mol))
        NC = EState._calculate_atom_estate(mol, atomic_num=6)
        NH = EState._calculate_atom_estate(mol, atomic_num=1)
        return Ntotal - NC - NH

    @staticmethod
    def calculate_average_estate(mol: Chem.Mol) -> float:
        """Calculate the ratio of the sum of the EState indices over heavy atoms and the number of non-hydrogen atoms."""
        N = mol.GetNumAtoms()
        return sum(EState._calculate_estate(mol)) / N

    @staticmethod
    def calculate_max_estate(mol: Chem.Mol) -> float:
        """Calculate the maximal Estate value in all atoms."""
        return max(EState._calculate_estate(mol))

    @staticmethod
    def calculate_min_estate(mol: Chem.Mol) -> float:
        """Calculate the minimal Estate value in all atoms."""
        return min(EState._calculate_estate(mol))

    @staticmethod
    def calculate_diff_max_min_estate(mol: Chem.Mol) -> float:
        """Calculate the difference between Smax and Smin."""
        return max(EState._calculate_estate(mol)) - min(EState._calculate_estate(mol))

    @staticmethod
    def calculate_estate_fingerprint(mol: Chem.Mol, implementation='rdkit', binary: bool = False) -> dict:
        """Calculate the sum of EState values for each EState atom type.

        :param implementation: either rdkit or chemopy. chemopy rounds
                               to the third decimal place but not rdkit.
        :param binary: should bineray values be returned instead of continous ones
        """
        if implementation not in ['rdkit', 'chemopy']:
            raise ValueError('Implementation of AtomTypeEState must be either rdkit or chemopy.')
        if implementation == 'chemopy':
            AT = ATEstate.get_atom_label(mol)
            Estate = EState._calculate_estate(mol)
            res = []
            for i in AT:
                if i == []:
                    res.append(0)
                else:
                    res.append(sum(Estate[k] for k in i))
            ESresult = {f'S{n + 1}': es for n, es in enumerate(res)}
            if binary:
                ESresult = {f'EStateFP_{i + 1}' : 0 if x == 0 else 1 for i, x in enumerate(ESresult.values())}
            return ESresult
        else:
            temp = ESFP.FingerprintMol(mol)
            if binary:
                res = {f'EStateFP_{i + 1}' : 0 if x == 0 else 1 for i, x in enumerate(temp[0])}
            else:
                res = {f'S{i + 1}': j for i, j in enumerate(temp[1])}
            return res

    @staticmethod
    def calculate_atom_type_estate_fingerprint(mol: Chem.Mol) -> dict:
        """Calculate EState Fingerprints.

        This is the counts of each EState atom type in the molecule.
        """
        temp = ESFP.FingerprintMol(mol)
        res = {}
        for i, j in enumerate(temp[0]):
            res[f'Sfinger{i + 1}'] = j
        return res

    @staticmethod
    def calculate_max_atom_type_estate(mol: Chem.Mol) -> dict:
        """Calculate the maximum of EState value."""
        AT = ATEstate.get_atom_label(mol)
        Estate = EState._calculate_estate(mol)
        res = []
        for i in AT:
            if i == []:
                res.append(0)
            else:
                res.append(max(Estate[k] for k in i))
        ESresult = {}
        for n, es in enumerate(res):
            ESresult[f'Smax{n + 1}'] = es
        return ESresult

    @staticmethod
    def calculate_min_atom_type_estate(mol: Chem.Mol) -> dict:
        """Calculate the minimum of EState value."""
        AT = ATEstate.get_atom_label(mol)
        Estate = EState._calculate_estate(mol)
        res = []
        for i in AT:
            if i == []:
                res.append(0)
            else:
                res.append(min(Estate[k] for k in i))
        ESresult = {}
        for n, es in enumerate(res):
            ESresult[f'Smin{n + 1}'] = es
        return ESresult

    @staticmethod
    def get_all_descriptors(mol: Chem.Mol) -> dict:
        """Calculate all (8) EState descriptors."""
        result = {}
        result.update({'Shev': EState.calculate_heavy_atom_estate(mol)})
        result.update({'Scar': EState.calculate_c_atom_estate(mol)})
        result.update({'Shal': EState.calculate_halogen_estate(mol)})
        result.update({'Shet': EState.calculate_hetero_estate(mol)})
        result.update({'Save': EState.calculate_average_estate(mol)})
        result.update({'Smax': EState.calculate_max_estate(mol)})
        result.update({'Smin': EState.calculate_min_estate(mol)})
        result.update({'DS': EState.calculate_diff_max_min_estate(mol)})
        return result

    @staticmethod
    def get_all_fps(mol: Chem.Mol) -> dict:
        """Calculate all (316) EState descriptors."""
        result = {}
        result.update(EState.calculate_estate_fingerprint(mol))
        result.update(EState.calculate_max_atom_type_estate(mol))
        result.update(EState.calculate_min_atom_type_estate(mol))
        return result



# -*- coding: utf-8 -*-


"""Moran autocorrelation descriptors."""

import numpy as np
from rdkit import Chem

from .atomproperty import get_relative_atomic_property


class Moran:
    """Moran autocorrelation descriptors."""

    @staticmethod
    def _calculate_moran_autocorrelation(mol: Chem.Mol, lag: int = 1, propertylabel: str = 'm') -> float:
        """Calculate weighted Moran autocorrelation descriptors.

        :param lag: topological distance between atom i and atom j.
        :param propertylabel: type of weighted property
        """
        Natom = mol.GetNumAtoms()
        prolist = []
        for i in mol.GetAtoms():
            temp = get_relative_atomic_property(i.GetSymbol(), propertyname=propertylabel)
            prolist.append(temp)
        aveweight = sum(prolist) / Natom
        tempp = [np.square(x - aveweight) for x in prolist]
        GetDistanceMatrix = Chem.GetDistanceMatrix(mol)
        res = 0.0
        index = 0
        for i in range(Natom):
            for j in range(Natom):
                if GetDistanceMatrix[i, j] == lag:
                    atom1 = mol.GetAtomWithIdx(i)
                    atom2 = mol.GetAtomWithIdx(j)
                    temp1 = get_relative_atomic_property(element=atom1.GetSymbol(), propertyname=propertylabel)
                    temp2 = get_relative_atomic_property(element=atom2.GetSymbol(), propertyname=propertylabel)
                    res = res + (temp1 - aveweight) * (temp2 - aveweight)
                    index += 1
                else:
                    res = res + 0.0
        if sum(tempp) == 0 or index == 0:
            result = 0
        else:
            result = (res / index) / (sum(tempp) / Natom)
        return result

    @staticmethod
    def calculate_moran_auto_mass(mol: Chem.Mol) -> dict:
        """Calculate Moran autocorrelation with carbon-scaled atomic mass."""
        res = {}
        for i in range(8):
            res[f'MATSm{i + 1}'] = Moran._calculate_moran_autocorrelation(mol, lag=i + 1, propertylabel='m')
        return res

    @staticmethod
    def calculate_moran_auto_volume(mol: Chem.Mol) -> dict:
        """Calculate Moran autocorrelation with carbon-scaled atomic van der Waals volume."""
        res = {}
        for i in range(8):
            res[f'MATSv{i + 1}'] = Moran._calculate_moran_autocorrelation(mol, lag=i + 1, propertylabel='V')
        return res

    @staticmethod
    def calculate_moran_auto_electronegativity(mol: Chem.Mol) -> dict:
        """Calculate Moran autocorrelation with carbon-scaled atomic Sanderson electronegativity."""
        res = {}
        for i in range(8):
            res[f'MATSe{i + 1}'] = Moran._calculate_moran_autocorrelation(mol, lag=i + 1, propertylabel='En')
        return res

    @staticmethod
    def calculate_moran_auto_polarizability(mol: Chem.Mol) -> dict:
        """Calculate Moran autocorrelation with carbon-scaled atomic polarizability."""
        res = {}
        for i in range(8):
            res[f'MATSp{i + 1}'] = Moran._calculate_moran_autocorrelation(mol, lag=i + 1, propertylabel='alapha')
        return res

    @staticmethod
    def get_all(mol: Chem.Mol) -> dict:
        """Calcualate all (32) Moran autocorrelation descriptors."""
        res = {}
        res.update(Moran.calculate_moran_auto_mass(mol))
        res.update(Moran.calculate_moran_auto_volume(mol))
        res.update(Moran.calculate_moran_auto_electronegativity(mol))
        res.update(Moran.calculate_moran_auto_polarizability(mol))
        return res



# -*- coding: utf-8 -*-


"""Charged partial surface area (CPSA) descriptors."""

from rdkit import Chem

from . import asa
from .geo_opt import get_atom_class_list, read_coordinates


class CPSA:
    """Charge partial surface area descriptors."""

    @staticmethod
    def get_charge_sa(arc_file, radius_probe=1.5, n_sphere_point=960):
        """Get atom symbol, charge and partial solvent-accessible surface areas for all atoms.

        :param arc_file: Path to MOPAC .arc file
        :param RadiusProbe: radius of the probe used to calculate SASA
        :param n_sphere_point: number of points per atom to calculate SASA
        """
        charge_coordinates = read_coordinates(arc_file)
        atoms = get_atom_class_list(charge_coordinates)
        fasa = asa.calculate_asa(atoms, radius_probe, n_sphere_point)
        res = []
        for i in range(len(fasa)):
            res.append([charge_coordinates[i][0], charge_coordinates[i][4], fasa[i]])
        return res
    
    @staticmethod
    def calculate_asa(charge_sa):
        """calculate solvent-accessible surface area."""
        res = 0.0
        for i in charge_sa:
            res = res + i[2]
        return res

    @staticmethod
    def calculate_msa(arc_file):
        """calculate molecular surface areas.

        :param arc_file: Path to MOPAC .arc file
        """
        charge_sa = CPSA.get_charge_sa(arc_file, radius_probe =0, n_sphere_point=960)
        res = 0.0
        for i in charge_sa:
            res = res + i[2]
        return res

    @staticmethod
    def calculate_pnsa1(charge_sa):
        """Calculate partial negative area."""
        res = 0.0
        for i in charge_sa:
            if float(i[1]) < 0:
                res = res + i[2]
        return res

    @staticmethod
    def calculate_ppsa1(charge_sa):
        """Calculate partial positive area."""
        res = 0.0
        for i in charge_sa:
            if float(i[1]) > 0:
                res = res + i[2]
        return res

    @staticmethod
    def calculate_pnsa2(charge_sa):
        """Calculate total charge weighted negative surface area."""
        temp1, temp2 = 0.0, 0.0
        for i in charge_sa:
            if float(i[1]) < 0:
                temp1 += float(i[1])
                temp2 += i[2]
        res = temp1 * temp2
        return res

    @staticmethod
    def calculate_ppsa2(charge_sa):
        """Calculate total charge weighted positive surface area."""
        temp1, temp2 = 0.0, 0.0
        for i in charge_sa:
            if float(i[1]) > 0:
                temp1 += float(i[1])
                temp2 += i[2]
        res = temp1 * temp2
        return res

    @staticmethod
    def calculate_pnsa3(charge_sa):
        """Calculate atom charge weighted negative surface area."""
        res = 0.0
        for i in charge_sa:
            if float(i[1]) < 0:
                res += float(i[1]) * i[2]
        return res

    @staticmethod
    def calculate_ppsa3(charge_sa):
        """Calculate atom charge weighted positive surface area."""
        res = 0.0
        for i in charge_sa:
            if float(i[1]) > 0:
                res += float(i[1]) * i[2]
        return res

    @staticmethod
    def calculate_dpsa1(charge_sa):
        """Calculate difference in charged partial surface area."""
        return CPSA.calculate_ppsa1(charge_sa) - CPSA.calculate_pnsa1(charge_sa)

    @staticmethod
    def calculate_dpsa2(charge_sa):
        """Calculate difference in total charge weighted partial surface area."""
        return CPSA.calculate_ppsa2(charge_sa) - CPSA.calculate_pnsa2(charge_sa)

    @staticmethod
    def calculate_dpsa3(charge_sa):
        """Calculate difference in atomic charge weighted surface area."""
        return CPSA.calculate_ppsa3(charge_sa) - CPSA.calculate_pnsa3(charge_sa)

    @staticmethod
    def calculate_fnsa1(charge_sa):
        """Calculate fractional charged partial negative surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_pnsa1(charge_sa) / temp

    @staticmethod
    def calculate_fnsa2(charge_sa):
        """Calculate fractional charged total negative surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_pnsa2(charge_sa) / temp

    @staticmethod
    def calculate_fnsa3(charge_sa):
        """Calculate fractional charged atom negative surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_pnsa3(charge_sa) / temp

    @staticmethod
    def calculate_fpsa1(charge_sa):
        """Calculate fractional charged partial positive surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_ppsa1(charge_sa) / temp

    @staticmethod
    def calculate_fpsa2(charge_sa):
        """Calculate fractional charged total positive surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_ppsa2(charge_sa) / temp

    @staticmethod
    def calculate_fpsa3(charge_sa):
        """Calculate fractional charged atom positive surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_ppsa3(charge_sa) / temp

    @staticmethod
    def calculate_wnsa1(charge_sa):
        """Calculate surface weighted charged partial negative surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_pnsa1(charge_sa) * temp / 1000

    @staticmethod
    def calculate_wnsa2(charge_sa) -> float:
        """Calculate surface weighted charged total negative surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_pnsa2(charge_sa) * temp / 1000

    @staticmethod
    def calculate_wnsa3(charge_sa) -> float:
        """Calculate surface weighted charged atom negative surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_pnsa3(charge_sa) * temp / 1000

    @staticmethod
    def calculate_wpsa1(charge_sa) -> float:
        """Calculate surface weighted charged partial positive surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_ppsa1(charge_sa) * temp / 1000

    @staticmethod
    def calculate_wpsa2(charge_sa) -> float:
        """Calculate surface weighted charged total positive surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_ppsa2(charge_sa) * temp / 1000

    @staticmethod
    def calculate_wpsa3(charge_sa) -> float:
        """Calculate surface weighted charged atom positive surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_ppsa3(charge_sa) * temp / 1000

    @staticmethod
    def calculate_tasa(charge_sa) -> float:
        """Calculate total apolar (hydrophobic) surface area."""
        res = 0.0
        for i in charge_sa:
            if abs(float(i[1])) < 0.2:
                res += i[2]
        return res

    @staticmethod
    def calculate_tpsa(charge_sa) -> float:
        """Calculate total polar surface area."""
        res = 0.0
        for i in charge_sa:
            if abs(float(i[1])) >= 0.2:
                res += i[2]
        return res

    @staticmethod
    def calculate_ratiotatp(charge_sa) -> float:
        """Calculate ratio between TASA and TPSA (FrTATP)."""
        res = 0.0
        if CPSA.calculate_tpsa(charge_sa) == 0:
            return res
        else:
            return CPSA.calculate_tasa(charge_sa) / CPSA.calculate_tpsa(charge_sa)

    @staticmethod
    def calculate_rasa(charge_sa) -> float:
        """Calculate relative hydrophobic surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_tasa(charge_sa) / temp

    @staticmethod
    def calculate_rpsa(charge_sa) -> float:
        """Calculate relative polar surface area."""
        temp = 0.0
        for i in charge_sa:
            temp += i[2]
        return CPSA.calculate_tpsa(charge_sa) / temp

    @staticmethod
    def calculate_rncs(charge_sa) -> float:
        """Calculate relative negative charge surface area."""
        charge = []
        for i in charge_sa:
            charge.append(float(i[1]))
        temp = []
        for i in charge_sa:
            temp.append(i[2])
        RNCG = min(charge) / sum(i for i in charge if i < 0)
        return temp[charge.index(min(charge))] / RNCG

    @staticmethod
    def calculate_rpcs(charge_sa) -> float:
        """Calculate relative positive charge surface area."""
        charge = []
        for i in charge_sa:
            charge.append(float(i[1]))
        temp = []
        for i in charge_sa:
            temp.append(i[2])
        RPCG = max(charge) / sum(i for i in charge if i > 0)
        return temp[charge.index(min(charge))] / RPCG

    @staticmethod
    def get_all(mol: Chem.Mol, arc_file: str) -> dict:
        """Get all CPSA descriptors.

        :param arc_file: Path to MOPAC .arc file
        """
        res = {}
        charge_sa = CPSA.get_charge_sa(arc_file, radius_probe=1.5, n_sphere_point=5000)
        res['SASA'] = CPSA.calculate_asa(charge_sa)
        res['MSA'] = CPSA.calculate_msa(arc_file)
        res['PNSA1'] = CPSA.calculate_pnsa1(charge_sa)
        res['PPSA1'] = CPSA.calculate_ppsa1(charge_sa)
        res['PNSA2'] = CPSA.calculate_pnsa2(charge_sa)
        res['PPSA2'] = CPSA.calculate_ppsa2(charge_sa)
        res['PNSA3'] = CPSA.calculate_pnsa3(charge_sa)
        res['PPSA3'] = CPSA.calculate_ppsa3(charge_sa)
        res['DPSA1'] = CPSA.calculate_dpsa1(charge_sa)
        res['DPSA2'] = CPSA.calculate_dpsa2(charge_sa)
        res['DPSA3'] = CPSA.calculate_dpsa3(charge_sa)
        res['FNSA1'] = CPSA.calculate_fnsa1(charge_sa)
        res['FNSA2'] = CPSA.calculate_fnsa2(charge_sa)
        res['FNSA3'] = CPSA.calculate_fnsa3(charge_sa)
        res['FPSA1'] = CPSA.calculate_fpsa1(charge_sa)
        res['FPSA2'] = CPSA.calculate_fpsa2(charge_sa)
        res['FPSA3'] = CPSA.calculate_fpsa3(charge_sa)
        res['WNSA1'] = CPSA.calculate_wnsa1(charge_sa)
        res['WNSA2'] = CPSA.calculate_wnsa2(charge_sa)
        res['WNSA3'] = CPSA.calculate_wnsa3(charge_sa)
        res['WPSA1'] = CPSA.calculate_wpsa1(charge_sa)
        res['WPSA2'] = CPSA.calculate_wpsa2(charge_sa)
        res['WPSA3'] = CPSA.calculate_wpsa3(charge_sa)
        res['TASA'] = CPSA.calculate_tasa(charge_sa)
        res['TotPSA'] = CPSA.calculate_tpsa(charge_sa)
        res['FrTATP'] = CPSA.calculate_ratiotatp(charge_sa)
        res['RASA'] = CPSA.calculate_rasa(charge_sa)
        res['RPSA'] = CPSA.calculate_rpsa(charge_sa)
        res['RNCS'] = CPSA.calculate_rncs(charge_sa)
        res['RPCS'] = CPSA.calculate_rpcs(charge_sa)
        return res



# -*- coding: utf-8 -*-


"""Molecule geometry optimization with MOPAC."""


import os
import shutil
import subprocess  # noqa: S404
import tempfile
import warnings
from sys import platform
from typing import List, Tuple, Union, Optional

from openbabel import pybel
from rdkit import Chem
from rdkit.rdBase import BlockLogs
from rdkit.Chem import rdDistGeom

from . import vector3d
from .utils import (get_file_in_dir_from_ext, get_lastest_created_file, is_in_subdirectory_tree)



# Default OpenMOPAC configuration
MOPAC_CONFIG = {'2016': ['mopac', ['PM7', 'PM6', 'PM3', 'AM1', 'MNDO']]}


# Disable openbabel logging
pybel.ob.obErrorLog.StopLogging()


class Atom:
    """Wrapper for atomic properties."""

    def __init__(self, Coordinates: List[float]):
        """Initialize an Atom object."""
        self.pos = vector3d.Vector3D()
        self.radius = 0.0
        self.coordinates = Coordinates
        self.element = ''

    def set_coordinates(self):
        """Parse raw ARC coordinates."""
        temp = self.coordinates
        self.pos.x = float(temp[1])
        self.pos.y = float(temp[2])
        self.pos.z = float(temp[3])

    def get_coordinates(self):
        """Get coordinates of the atom."""
        self.set_coordinates()
        return self.pos

    def set_element(self):
        """Set element from raw ARC coordinates."""
        temp = self.coordinates
        self.element = temp[0]

    def get_element(self):
        """Get element."""
        self.set_element()
        return self.element

    def set_radius(self):
        """Set radius."""
        radii = {'H': 1.20, 'N': 1.55, 'Na': 2.27, 'Cu': 1.40, 'Cl': 1.75, 'C': 1.70,
                 'O': 1.52, 'I': 1.98, 'P': 1.80, 'B': 1.85, 'Br': 1.85, 'S': 1.80, 'Se': 1.90,
                 'F': 1.47, 'Fe': 1.80, 'K': 2.75, 'Mn': 1.73, 'Mg': 1.73, 'Zn': 1.39, 'Hg': 1.8,
                 'Li': 1.8, '.': 1.8}
        temp = self.get_element()
        if temp in radii.keys():
            self.radius = radii[temp]
        else:
            self.radius = radii['.']

    def get_radius(self):
        """Get the radius of the atom."""
        self.set_radius()
        return self.radius


def get_atom_class_list(coordinates: List[List[float]]) -> List[Atom]:
    """Get a list of atoms from a list of raw ARC coordinates.

    :param coordinates: raw ARC coordinates as returned by _ReadCoordinates
    """
    atoms = []
    for coords in coordinates:
        atom = Atom(coords)
        atom.set_coordinates()
        atom.set_element()
        atom.set_radius()
        atoms.append(atom)
    return atoms


def read_coordinates(arc_file: str):
    """Read coordinates and charges of atoms from a MOPAC ARC file.

    :param arc_file: Path to MOPAC .arc file
    """
    res = []
    with open(arc_file, 'r') as f:
        templine = f.readlines()
    for line in range(len(templine)):
        if templine[line][-7: -1] == "CHARGE":
            k = line
            break
    for i in templine[k + 4: len(templine) - 1]:
        temp = i.split()
        ElementCoordinate = [temp[0].strip(), temp[1].strip(),
                             temp[3].strip(), temp[5].strip(),
                             temp[-1].strip()]
        res.append(ElementCoordinate)
    return res


class MopacInputFile:
    """File holding inputs for MOPAC molecular geometry optimization."""

    periodic_table = Chem.GetPeriodicTable()

    def __init__(self, path:str, method='PM7', version='2016'):
        """Create an instance of a MOPAC input file."""
        self.path = path
        self.method = method
        self.version = version

    def open(self):
        """Open the handle"""
        self.handle = open(self.path, 'w')
        self.handle.write(self.method + (' PRTCHAR ' if self.version == '2016' else ''))

    def write(self, mol: Chem.Mol, conf_id: int = -1):
        """Write a molecule to the input MOPAC file.

        :param mol: RDKit molecule to write
        :param conf_id: conformer id
        """
        # If molecule is charged
        charge = Chem.GetFormalCharge(mol)
        self.handle.write(f" CHARGE={'+' if charge > 0 else '-'}{charge}" if charge != 0 else '')
        self.handle.write('\n\n\n')
        # Write atomic positions and set them all for optimization
        conf = mol.GetConformer(conf_id)
        for atom in mol.GetAtoms():
            index = atom.GetIdx()
            elem = self.periodic_table.GetElementSymbol(atom.GetAtomicNum())
            pos = conf.GetAtomPosition(index)
            self.handle.write(f'{elem: <3}{pos.x:8.5f} 1 {pos.y:8.5f} 1 {pos.z:8.5f} 1\n')

    def close(self):
        """Close the internal handle."""
        self.handle.close()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class MopacResultDir:
    """Directory holding all inputs and outputs of MOPAC."""

    def __init__(self, path: Optional[str] = None):
        """Create an instance of a MOPAC input/output directory.

        :param exists: whether the directory already exists
        """
        if path is not None:
            if not os.path.isdir(path):
                raise ValueError('Path should either be specified if the directory already exists or left blank.')
        self.already_exists = path is not None
        self.path = path

    def open(self):
        if self.path is None:
            self.path = tempfile.mkdtemp()
        return self

    def close(self):
        if not self.already_exists:
            clean_mopac_files(self.path)

    def __enter__(self):
        return self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__del__()

    def __del__(self):
        self.close()


def format_conversion(inputmol: Chem.Mol,
                      method='PM7', version='2016',
                      outfile=None, outdir=None,
                      ) -> Tuple[MopacResultDir, str]:
    """Prepare a molecule to be optimized by MOPAC.

    The preparation goes as follows:

    1. all hydrogens are made explicit,
    2. a 3D conformer is generated from RDKit's stochastic search,
       based on distance geometry and  exerimental crystallographic knowledge,
       Wang S., J. Chem. Inf. Model. (2020), 60(4),2044-2058.
    3. the conformer is converted to a first MOPAC input file with OpenBabel,
    4. semi-empirical method and MOPAC version is added to the MOPAC input file.

    :param method: MOPAC semi-empirical method to be used for molecular geometry oprimization
    :param version: version of MOPAC to use
    :param outfile: name of the output the MOPAC input file
    :param outdir: directory where to create the MOPAC input file
        If not specified, a temporary directory is created but
        THE USER MUST REMOVE IT THEMSELVES USING geo_opt.clean_mopac_files.
    :return: the directory where the MOPAC input file was created
    """
    # Ensure method and MOPAC version are compatible
    if not is_method_supported_by_mopac(method, version):
        raise ValueError(f'Method {method} is not supported by MOPAC {version}.')
    # Step 1: add Hs
    # Step 2: generate conformer
    confs = list(inputmol.GetConformers())
    if not (len(confs) > 0 and confs[-1].Is3D()):
        inputmol = Chem.AddHs(inputmol)
        rdDistGeom.EmbedMolecule(inputmol, rdDistGeom.ETKDGv3())
    else:
        inputmol = Chem.AddHs(inputmol, addCoords=True)
    running_dir = MopacResultDir().open() if outdir is None else outdir
    mpo_name = 'temp' if outfile is None else outfile
    # Step 3: create MOPAC input file
    with MopacInputFile(os.path.join(running_dir.path, f'{mpo_name}.dat'), method=method, version=version) as output_file:
        output_file.write(inputmol)
    return running_dir, f'{mpo_name}.dat'


def run_mopac(filename: str, version: str = '2016') -> int:
    """Run the MOPAC on a well-prepared input file.

    Parse default MOPAC config file if not read already.

    :param filename: path to the well-prepared MOPAC input file
    :param version: MOPAC version to be used
    """
    # Ensure all requirements are set
    if not is_mopac_version_available(version):
        raise ValueError(f'MOPAC version {version} is not available. Check your MOPAC config file.')
    # Run optimization
    mopac_bin = MOPAC_CONFIG[str(version)][0]
    try:
        retcode = subprocess.call(f'{mopac_bin} {filename}', shell=False,
                                  stdin=subprocess.DEVNULL,
                                  stdout=subprocess.DEVNULL,
                                  stderr=subprocess.DEVNULL)  # noqa: S603
        return retcode
    except Exception:
        return 1


def clean_mopac_files(dir_: str, force: bool = False) -> None:
    """Properly dispose of a temporary folder.

    :param force: whether to allow removing a non-temporary directory.
    """
    # If neither is dir_ in tempdir nor force set to True
    if not (is_in_subdirectory_tree(dir_, tempfile.gettempdir()) or force):
        raise PermissionError('Directory is not temporary. If you know what you '
                              'are doing force the  deletion')
    shutil.rmtree(dir_, ignore_errors=True)


def is_mopac_version_available(version: str) -> bool:
    """Return if the desired version of MOPAC can be used.

    :param version: version of MOPAC
    """
    # Is version configured?
    if version not in MOPAC_CONFIG.keys():
        return False
    # Does the executable/environment variable exist?
    if shutil.which(MOPAC_CONFIG[version][0]) is None:
        return False
    # Ensure the executable runs without trouble
    if platform.startswith('win32'):
        echo_cmd = 'echo.'
    elif platform in ('linux', 'darwin'):
        echo_cmd = r'echo -e "\n"'
    else:
        raise RuntimeError(f'Platform ({platform}) not supported.')
    retcode = subprocess.call(f'{echo_cmd} | {MOPAC_CONFIG[version][0]}', shell=True,
                              stdin=subprocess.DEVNULL,
                              stdout=subprocess.DEVNULL,
                              stderr=subprocess.DEVNULL)
    return retcode == 0


def is_method_supported_by_mopac(method: str, version: str) -> bool:
    """Return if the version of MOPAC supports a specific method.

    :param method: semi-empirical method to be applied
    :param version: version of MOPAC
    """
    if not is_mopac_version_available(version):
        raise ValueError(f'MOPAC version {version} is not available.')
    return str(method) in MOPAC_CONFIG[f'{version}'][1]


def get_arc_file(inputmol: Chem.Mol, method: str = 'PM7', version: str = '2016',
                 verbose: bool = True, exit_on_fail: bool = False,
                 ) -> Union[Tuple[MopacResultDir, str], None]:
    """Optimize molecule geometry with MOPAC.

    :param inputmol: molecule to optimize
    :param method: semi-empirical method to apply
    :param version: version of MOPAC to be used
    :param verbose: whether to print progress messages
    :param exit_on_fail: if False, if a method fails at generating
                         a structure, others are tried from most to
                         least accurate.
                         if True, return False on failure.
    :return: Tuple of (path to folder, path to arc_file) on success,
             None otherwise.
    """
    # Create proper input file
    dir_, dat_file = format_conversion(inputmol, method, version)
    # Ensure dat file exists
    full_path = os.path.join(dir_.path, dat_file)
    if not os.path.isfile(full_path):
        raise FileNotFoundError('Molecule could not be prepared for MOPAC.')
    # Run MOPAC
    retcode = run_mopac(full_path, version)
    # Get generated file
    # Different versions of MOPAC handle the outputname differently
    # e.g. 7.1 appends .arc after the .dat giving a .dat.arc file
    # while 2016 replaces the .dat by .arc
    output = get_file_in_dir_from_ext(dir_.path, '.arc')
    # Success when return code is 0
    success = not retcode and len(output) > 0
    if success:
        if verbose:
            print(f'Molecule geometry was successfully optimized using {method}.')  # noqa T001
        output = output[0] if len(output) == 1 else get_lastest_created_file(filepaths=output)
        # Rename output
        curated_filename = f'{os.path.splitext(full_path)[0]}.arc'
        os.rename(output, curated_filename)
        return dir_, curated_filename
    elif exit_on_fail:
        if verbose:
            print(f'Geometry optimization failed with {method}')  # noqa T001
        return
    else:  # neither success nor exit_on_fail
        if verbose:
            print(f'Geometry optimization failed with {method}')  # noqa T001
        methods_tried = MOPAC_CONFIG[f'{version}'][1]
        # Remove the method that was just used
        methods_tried.remove(method)
        # Try all possible methods, from most to least accurate
        skip_first_time = True
        while not success:
            if not skip_first_time:
                del methods_tried[0]  # Remove method tried to allow going to the next
                if len(methods_tried) == 0:  # No method left
                    if verbose:
                        print('Molecule could not be optimized.')  # noqa T001
                    else:
                        warnings.warn('Molecule could not be optimized', RuntimeWarning)
                    return
            skip_first_time = False
            if verbose:
                print(f'Attempting optimization with {methods_tried[0]}')  # noqa T001
            # Create proper input file
            new_attempt = get_arc_file(inputmol, methods_tried[0], version, verbose, exit_on_fail=True)
            if new_attempt is not None:
                return new_attempt


def get_optimized_mol(arc_file: str = None, inputmol: Chem.Mol = None,
                      method: str = 'PM7', version: str = '2016',
                      verbose: bool = False, dispose: bool = True,
                      ) -> Union[Chem.Mol, None]:
    """Optimize molecule geometry with MOPAC and return the optimized molecule.

    In order not to optimize a molecule multiple times, an ARC file may be provided.
    The path to the MOPAC output file is determined based on the ARC file provided.

    If not already optimized, a molecule may be provided.

    :param arc_file: Path to MOPAC .arc file
                     (ignored if inputmol provided).
    :param inputmol: molecule to optimize
                     (ignored if arc_file provided).
    :param method: semi-empirical method to apply
                   (ignored if arc_file provided).
    :param version: version of MOPAC to be used
                    (ignored if arc_file provided).
    :param verbose: whether to print progress messages
                    (ignored if arc_file provided).
    :param dispose: whether to remove generated MOPAC output files
                    (ignored if arc_file provided).
    :return: optimized rdkit molecule on success, None otherwise.
    """
    if arc_file is None and inputmol is None:
        raise ValueError('Either ARC file or inputmolecule must be provided.')
    if arc_file is not None:
        mopac_out_dir = os.path.dirname(arc_file)
        mopac_out_path = get_file_in_dir_from_ext(mopac_out_dir, '.out')
        mopac_out_path = mopac_out_path[0] if len(mopac_out_path) == 1 \
            else get_lastest_created_file(filepaths=mopac_out_path)
        if not len(mopac_out_path):
            return None
        pybelmol = next(pybel.readfile('mopout', mopac_out_path))
        # Convert to RDKit
        block = BlockLogs()
        mol = Chem.MolFromMol2Block(pybelmol.write(format='mol2'))
        if mol is None:
            mol = Chem.MolFromMolBlock(pybelmol.write(format='sdf'))
        del block
        return mol
    else:
        res = get_arc_file(inputmol, method, version, verbose, False)
        if res is None:
            return None
        dir_, arc_file_ = res
        mopac_out_path = get_file_in_dir_from_ext(dir_.path, '.out')
        mopac_out_path = mopac_out_path[0] if len(mopac_out_path) == 1 \
            else get_lastest_created_file(filepaths=mopac_out_path)
        if not len(mopac_out_path):
            return None
        pybelmol = next(pybel.readfile('mopout', mopac_out_path))
        if dispose:
            dir_.close()
        # Convert to RDKit
        block = BlockLogs()
        mol = Chem.MolFromMol2Block(pybelmol.write(format='mol2'), removeHs=False)
        if mol is None:
            mol = Chem.MolFromMolBlock(pybelmol.write(format='sdf'))
        del block
        return mol



# -*- coding: utf-8 -*-


"""3D MoRSE descriptors."""


import math
from typing import List

import numpy as np
from rdkit import Chem

from .atomproperty import get_relative_atomic_property
from .geo_opt import read_coordinates
from .utils import get_geometrical_distance_matrix, get_r

# Parameter for RDF equation
_beta = 100

class MoRSE:
    """3D MoRSE descriptors."""

    @staticmethod
    def calculate_unweight_morse(charge_coordinates: List[List[float]]) -> dict:
        """Calculate unweighted 3D MoRse descriptors.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        R = get_r(n=30)
        temp = []
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + math.sin(Ri * DM[j, k]) / (Ri * DM[j, k])
            RDFresult[f'MoRSEU{kkk + 1}'] = res
        return RDFresult

    @staticmethod
    def calculate_charge_morse(charge_coordinates: List[List[float]]) -> dict:
        """Calculate 3D MoRse descriptors from atomic charge.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        R = get_r(n=30)
        temp = []
        charge = []
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
            charge.append(float(i[4]))
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + charge[j] * charge[k] * math.sin(Ri * DM[j, k]) / (Ri * DM[j, k])
            RDFresult[f'MoRSEC{kkk + 1}'] = res
        return RDFresult

    @staticmethod
    def calculate_mass_morse(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> dict:
        """Calculate 3D MoRse descriptors from atomic mass.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        R = get_r(n=30)
        temp = []
        mass = [atom.GetMass() for atom in Chem.AddHs(mol).GetAtoms()]
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res += mass[j] * mass[k] * math.sin(Ri * DM[j, k]) / (Ri * DM[j, k])
            RDFresult[f'MoRSEM{kkk + 1}'] = res / 144
        return RDFresult

    @staticmethod
    def calculate_atomic_number_morse(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> dict:
        """Calculate 3D MoRse descriptors from atomic number.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        R = get_r(n=30)
        temp = []
        mass = [atom.GetMass() for atom in Chem.AddHs(mol).GetAtoms()]
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res += mass[j] * mass[k] * math.sin(Ri * DM[j, k]) / (Ri * DM[j, k])
            RDFresult[f'MoRSEN{kkk + 1}'] = res / 144
        return RDFresult

    @staticmethod
    def calculate_polarizability_morse(charge_coordinates: List[List[float]]) -> dict:
        """Calculate 3D MoRse descriptors from atomic polarizablity.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        R = get_r(n=30)
        temp = []
        polarizability = []
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
            polarizability.append(get_relative_atomic_property(i[0], 'alapha'))
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + polarizability[j] * polarizability[k] * math.sin(Ri * DM[j, k]) / (Ri * DM[j, k])
            RDFresult[f'MoRSEP{kkk + 1}'] = res
        return RDFresult

    @staticmethod
    def calculate_sanderson_electronegativity_morse(charge_coordinates: List[List[float]]) -> dict:
        """Calculate 3D MoRse descriptors from Sanderson electronegativity.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        R = get_r(n=30)
        temp = []
        En = []
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
            En.append(get_relative_atomic_property(i[0], 'En'))
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + En[j] * En[k] * math.sin(Ri * DM[j, k]) / (Ri * DM[j, k])
            RDFresult[f'MoRSEE{kkk + 1}'] = res
        return RDFresult

    @staticmethod
    def calculate_vdw_volume_morse(charge_coordinates: List[List[float]]) -> dict:
        """Calculate 3D MoRse descriptors from van der Waals volume.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        R = get_r(n=30)
        temp = []
        VDW = []
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
            VDW.append(get_relative_atomic_property(i[0], 'V'))
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + VDW[j] * VDW[k] * math.sin(Ri * DM[j, k]) / (Ri * DM[j, k])
            RDFresult[f'MoRSEV{kkk + 1}'] = res
        return RDFresult

    @staticmethod
    def get_morse_unweighted(arc_file: str) -> dict:
        """Get all unweighted 3D-Morse descriptors.

        :param arc_file: Path to MOPAC .arc file
        """
        ChargeCoordinates = read_coordinates(arc_file)
        result = MoRSE.CalculateUnweightMoRSE(ChargeCoordinates)
        return result

    @staticmethod
    def get_morse_charge(arc_file: str) -> dict:
        """Get all 3D-Morse descriptors from charge schemes.

        :param arc_file: Path to MOPAC .arc file
        """
        ChargeCoordinates = read_coordinates(arc_file)
        result = MoRSE.CalculateChargeMoRSE(ChargeCoordinates)
        return result

    @staticmethod
    def get_morse_mass(mol: Chem.Mol, arc_file: str) -> dict:
        """Get all 3D-Morse descriptors from on mass schemes.

        :param arc_file: Path to MOPAC .arc file
        """
        ChargeCoordinates = read_coordinates(arc_file)
        result = MoRSE.CalculateMassMoRSE(mol, ChargeCoordinates)
        return result

    @staticmethod
    def get_morse_atomic_number(mol: Chem.Mol, arc_file: str) -> dict:
        """Get all 3D-Morse descriptors from atomic number schemes.

        :param arc_file: Path to MOPAC .arc file
        """
        ChargeCoordinates = read_coordinates(arc_file)
        result = MoRSE.CalculateAtomicNumberMoRSE(mol, ChargeCoordinates)
        return result

    @staticmethod
    def get_morse_polarizability(arc_file: str) -> dict:
        """Get all 3D-Morse descriptors from polarizability schemes.

        :param arc_file: Path to MOPAC .arc file
        """
        ChargeCoordinates = read_coordinates(arc_file)
        result = MoRSE.CalculatePolarizabilityMoRSE(ChargeCoordinates)
        return result

    @staticmethod
    def get_morse_sanderson_electronegativity(arc_file: str) -> dict:
        """Get all 3D-Morse descriptors from Sanderson Electronegativity schemes.

        :param arc_file: Path to MOPAC .arc file
        """
        ChargeCoordinates = read_coordinates(arc_file)
        result = MoRSE.CalculateSandersonElectronegativityMoRSE(ChargeCoordinates)
        return result

    @staticmethod
    def get_morse_vdw_volume(arc_file: str) -> dict:
        """Get all 3D-Morse descriptors from VDW Volume schemes.

        :param arc_file: Path to MOPAC .arc file
        """
        ChargeCoordinates = read_coordinates(arc_file)
        result = MoRSE.CalculateVDWVolumeMoRSE(ChargeCoordinates)
        return result

    @staticmethod
    def get_all(mol: Chem.Mol, arc_file: str) -> dict:
        """Get all (210) 3D-Morse descriptors with different (un)weighted schemes.

        :param arc_file: Path to MOPAC .arc file
        """
        result = {}
        charge_coordinates = read_coordinates(arc_file)
        result.update(MoRSE.calculate_unweight_morse(charge_coordinates))
        result.update(MoRSE.calculate_charge_morse(charge_coordinates))
        result.update(MoRSE.calculate_mass_morse(mol, charge_coordinates)) if mol is not None else np.nan
        result.update(MoRSE.calculate_atomic_number_morse(mol, charge_coordinates)) if mol is not None else np.nan
        result.update(MoRSE.calculate_polarizability_morse(charge_coordinates))
        result.update(MoRSE.calculate_sanderson_electronegativity_morse(charge_coordinates))
        result.update(MoRSE.calculate_vdw_volume_morse(charge_coordinates))
        return result



# -*- coding: utf-8 -*-


"""Routines to calculate the Accessible Surface Area of a set of atoms.

The algorithm is adapted from the Rose lab's chasa.py, which uses
the dot density technique found in:

Shrake, A., and J. A. Rupley. "Environment and Exposure to Solvent
of Protein Atoms. Lysozyme and Insulin." JMB (1973) 79:351-371.
"""


import math
from typing import List

from .geo_opt import Atom
from .vector3d import Vector3D, pos_distance, pos_distance_sq


def generate_sphere_points(n: int) -> List[Vector3D]:
    """Distribute points on a sphere using the Golden Section Spiral algorithm.

    :param n: number of points to generate along the sphere surface.
    """
    points = []
    inc = math.pi * (3 - math.sqrt(5))
    offset = 2 / float(n)
    for k in range(int(n)):
        y = k * offset - 1 + (offset / 2)
        r = math.sqrt(1 - y * y)
        phi = k * inc
        points.append(Vector3D(math.cos(phi) * r, y, math.sin(phi) * r))
        # points.append([math.cos(phi) * r, y, math.sin(phi) * r])
    return points


def find_neighbor_indices(atoms: List[Atom], probe: float, k: int) -> List[int]:
    """Return indices of atoms within probe distance to atom k.

    If another atom u is found to be distant from atom k by less than
    the sum of its radius, the radius of atom k, and the diameter of
    the probe, then the probe cannot fit in between those and atom u.
    Hnce atom u is considered a neighbour of atom k.

    :param atoms: list of atoms
    :param probe: radius of probe
    :param k: atom from which the search is performed.
    """
    neighbor_indices = []
    atom_k = atoms[k]
    radius = atom_k.radius + 2 * probe
    indices = list(range(k))
    indices.extend(range(k + 1, len(atoms)))
    for i in indices:
        atom_i = atoms[i]
        dist = pos_distance(atom_k.pos, atom_i.pos)
        if dist < radius + atom_i.radius:
            neighbor_indices.append(i)
    return neighbor_indices


def calculate_asa(atoms: List[Atom], probe: float, n_sphere_point: int = 960) -> List[float]:
    """Return the accessible surface area of the atoms.

    :param atoms: list of atoms to get the surface area of
    :param probe: radius of the probe
    :param n_sphere_point: number of evenly distributed
                           points along the atoms' surface
    """
    sphere_points = generate_sphere_points(n_sphere_point)

    const = 4.0 * math.pi / len(sphere_points)
    test_point = Vector3D()
    areas = []
    for i, atom_i in enumerate(atoms):

        neighbor_indices = find_neighbor_indices(atoms, probe, i)
        n_neighbor = len(neighbor_indices)
        j_closest_neighbor = 0
        radius = probe + atom_i.radius

        n_accessible_point = 0
        for point in sphere_points:
            is_accessible = True

            test_point.x = point.x * radius + atom_i.pos.x
            test_point.y = point.y * radius + atom_i.pos.y
            test_point.z = point.z * radius + atom_i.pos.z

            cycled_indices = list(range(j_closest_neighbor, n_neighbor))
            cycled_indices.extend(range(j_closest_neighbor))

            for j in cycled_indices:
                atom_j = atoms[neighbor_indices[j]]
                r = atom_j.radius + probe
                diff_sq = pos_distance_sq(atom_j.pos, test_point)
                if diff_sq < r * r:
                    j_closest_neighbor = j
                    is_accessible = False
                    break
            if is_accessible:
                n_accessible_point += 1

        area = const * n_accessible_point * radius * radius
        areas.append(area)
    return areas



# -*- coding: utf-8 -*-


"""Multiple 3D molecular fingerprints."""

from typing import Union, List, Optional

from e3fp.fingerprint.fprinter import Fingerprinter
from rdkit import Chem

from .geo_opt import get_optimized_mol


class Fingerprint3D:
    """Molecular 3D fingerprints."""

    @staticmethod
    def calculate_e3fp(mol: Union[Chem.Mol, List[Chem.Mol]], arc_file: str = None, nbits: int = 2048, level: int = 5,
                       radius_multiplier: float = 1.718, stereo: bool = True, counts: bool = False,
                       include_disconnected: bool = True, rdkit_invariants: bool = False,
                       exclude_floating: bool = True, remove_duplicate_substructs: bool = True) -> Union[dict, List[dict]]:
        """Calculate extended 3D fingerprint(s).

        :param mol: MOPAC optimized molecule (result of `geo_opt.get_optimized_mol`)
        :param arc_file: Ignored
        :param nbits: number of bits in the folded fingerprint
        :param level: maximum number of iterations for fingerprint generation
        :param radius_multiplier: ratio by which to increase shell size
        :param stereo: differentiate based on stereography
        :param counts: should a count fingerprint be generated
        :param include_disconnected: include disconnected atoms from hashes and substructure
        :param rdkit_invariants: use the atom invariants used by RDKit for its Morgan fingerprint
        :param exclude_floating: exclude atoms with no bonds
        :param remove_duplicate_substructs: drop duplicate hashes in the fingerprint
        :return: a dictionary if only one molecule is provided, otherwise a list of dicts
        """
        fper = Fingerprinter(bits=nbits, level=level,radius_multiplier=radius_multiplier,
                             stereo=stereo, counts=counts, include_disconnected=include_disconnected,
                             rdkit_invariants=rdkit_invariants, exclude_floating=exclude_floating,
                             remove_duplicate_substructs=remove_duplicate_substructs)
        # Convenience lambda function
        to_dense_fp = lambda bits, size: [1 if i in bits else 0 for i in range(size)]
        # If unique molecule transform to list
        if not isinstance(mol, list):
            mol = [mol]
        # Calculate fingerprint
        fps = []
        for mol_ in mol:
            # Ensure conformers are available, optimize otherwise
            confs = list(mol_.GetConformers())
            if not (len(confs) > 0 and confs[-1].Is3D()):
                mol_ = get_optimized_mol(inputmol=mol_)
            fper.run(mol=mol_)
            fp = to_dense_fp(fper.get_fingerprint_at_level().indices, nbits)
            values = dict(zip([f'E3FP_{i + 1}' for i in range(nbits)],
                              fp))
            fps.append(values)
        # Return only one value if input was not a list
        if len(mol) == 1:
            return fps[0]
        return fps

    _fp_funcs = {'E3FP': (calculate_e3fp, ('nbits')),
                 }

    @staticmethod
    def get_all_fps(mol: Chem.Mol, nbits: Optional[int] = None) -> dict:
        """Calculate all fingerprints."""
        values = {}
        for des_label, (func, supported_args) in Fingerprint3D._fp_funcs.items():
            if nbits is not None and 'nbits' in supported_args:
                values.update(func(mol, nbits=nbits))
            else:
                values.update(func(mol))
        return values
